__all__ = ['test_cloud_rain_Memmap',
           'test_height_pressure_Memmap',
           'test_height_pressure_Read',
           'test_humidity_Memmap',
           'test_humidity_Read',
           'test_ipr_Memmap',
           'test_ipr_Read',
           'test_irr_Read',
           'test_irr_Memmap',
           'test_landuse_Memmap',
           'test_lateral_boundary_Memmap',
           'test_one3d_Memmap',
           'test_one3d_Read',
           'test_point_source_Memmap',
           'test_point_source_Read',
           'test_temperature_Memmap',
           'test_temperature_Read',
           'test_uamiv_Memmap',
           'test_uamiv_Read',
           'test_uamiv_Write',
           'test_vertical_diffusivity_Memmap',
           'test_vertical_diffusivity_Read',
           'test_wind_Memmap',
           'test_wind_Read']

import PseudoNetCDF.camxfiles as cx
test_cloud_rain_Memmap = cx.cloud_rain.Memmap.TestMemmap
# test_finst_Memmap = cx.finst.Memmap.TestMemmap
test_height_pressure_Memmap = cx.height_pressure.Memmap.TestMemmap
test_height_pressure_Read = cx.height_pressure.Read.TestRead
test_humidity_Memmap = cx.humidity.Memmap.TestMemmap
test_humidity_Read = cx.humidity.Read.TestRead
test_ipr_Memmap = cx.ipr.Memmap.TestMemmap
test_ipr_Read = cx.ipr.Read.TestRead
test_irr_Read = cx.irr.Read.TestRead
test_irr_Memmap = cx.irr.Memmap.TestMemmap
test_landuse_Memmap = cx.landuse.Memmap.TestMemmap
test_lateral_boundary_Memmap = cx.lateral_boundary.Memmap.TestMemmap
test_one3d_Memmap = cx.one3d.Memmap.TestMemmap
test_one3d_Read = cx.one3d.Read.TestRead
test_point_source_Memmap = cx.point_source.Memmap.TestMemmap
test_point_source_Read = cx.point_source.Read.TestRead
test_temperature_Memmap = cx.temperature.Memmap.TestMemmap
test_temperature_Read = cx.temperature.Read.TestRead
test_uamiv_Memmap = cx.uamiv.Memmap.TestMemmap
test_uamiv_Read = cx.uamiv.Read.TestuamivRead
test_uamiv_Write = cx.uamiv.Write.TestMemmaps
test_vertical_diffusivity_Memmap = cx.vertical_diffusivity.Memmap.TestMemmap
test_vertical_diffusivity_Read = cx.vertical_diffusivity.Read.TestRead
test_wind_Memmap = cx.wind.Memmap.TestMemmap
test_wind_Read = cx.wind.Read.TestRead



from __future__ import print_function
from PseudoNetCDF import PseudoNetCDFFile, PseudoNetCDFVariable
import re
import numpy as np


class jtable(PseudoNetCDFFile):
    def __init__(self, path):
        self.dimensions = {}
        self.variables = {}
        fobj = open(path)
        jdate = int(fobj.readline().split()[0])
        nlevels = int(fobj.readline().split()[0])
        levels = [float(x) for x in fobj.readline().split()]
        nlats = int(fobj.readline().split()[0])
        lats = [float(x) for x in fobj.readline().split()]
        nangles = int(fobj.readline().split()[0])
        angles = [float(x) for x in fobj.readline().split()]
        nrxns = int(fobj.readline().split()[0])
        scaling = dict()
        rxns = []
        for ri in range(nrxns):
            tname, scale = fobj.readline().split(',')
            tname = tname.replace("'", "").strip()
            scaling[tname] = eval(scale.strip())
            rxns.append(tname)
        lines = fobj.readlines()
        header = re.compile('\s+\d+\s+\d+\s+\d+\s*')
        data = []
        headers = []
        for line in lines:
            if header.match(line):
                headers.append(line)
            else:
                data.append(line)
        data = np.fromstring(''.join(data), sep=' ').reshape(
            nlevels, nlats, nrxns, nangles)
        headers = np.fromstring(''.join(headers), sep=' ').reshape(
            nlevels, nlats, nrxns, 3)
        assert((np.diff(headers[..., 0], axis=0) == 1).all())
        assert((np.diff(headers[..., 1], axis=1) == 1).all())
        assert((np.diff(headers[..., 2], axis=2) == 1).all())
        self.SDATE = jdate
        self.NLAYS = nlevels
        self.NLATS = nlats
        self.NANGLES = nangles
        self.NRXNS = nrxns
        self.createDimension('LAY', nlevels)
        self.variables['LAY'] = PseudoNetCDFVariable(
            self, 'LAY', 'f', ('LAY',),
            values=np.array(levels, dtype='f'), units='m')
        self.createDimension('LAT', nlats)
        self.variables['LAT'] = PseudoNetCDFVariable(
            self, 'LAT', 'f', ('LAT',), values=np.array(lats, dtype='f'),
            units='degrees')
        self.createDimension('ANGLE', nangles)
        self.variables['ANGLE'] = PseudoNetCDFVariable(
            self, 'ANGLE', 'f', ('ANGLE',), values=np.array(angles, dtype='f'),
            units='hours from noon')

        for rxni, rxn in enumerate(rxns):
            self.variables[rxn] = PseudoNetCDFVariable(
                self, rxn, 'f', ('LAY', 'LAT', 'ANGLE'),
                values=data[..., rxni, :].copy(), units='s**-1')

    def interpj(self, key_or_var, lay, lat, angle):
        from scipy.interpolate import griddata
        lays = self.variables['LAY']
        lats = self.variables['LAT']
        angles = self.variables['ANGLE']
        points = np.array([[[(lay_, lat_, angle_) for angle_ in angles]
                            for lat_ in lats] for lay_ in lays]).reshape(-1, 3)
        if isinstance(key_or_var, str):
            v = self.variables[key_or_var].ravel()
        else:
            v = np.array(key_or_var)
            assert(key_or_var.shape == (self.NLAYS, self.NLATS, self.NANGLES))
        if isinstance(lay, (int, float)):
            idx = np.array((lay, lat, angle), ndmin=2)
        else:
            idx = np.array((lay, lat, angle)).swapaxes(0, 1)
            assert(idx.shape[0] == 2)
        return griddata(points, v, idx)[0]


if __name__ == '__main__':
    j = jtable(path='/Volumes/LaCie/JTABLE_1985172')
    v = j.variables['HCHO_M_SAPRC99']
    np.set_printoptions(precision=5)
    print(v[:2, :2, :2])
    print(j.interpj('HCHO_M_SAPRC99', 0, 10, 0))
    print(j.interpj('HCHO_M_SAPRC99', 0, 10, 1))
    print(j.interpj('HCHO_M_SAPRC99', 0, 20, 1))
    print(j.interpj('HCHO_M_SAPRC99', 1000, 20, 1))
    print(j.interpj('HCHO_M_SAPRC99', 0, 10, .5))
    print(j.interpj('HCHO_M_SAPRC99', 0, 15, .5))
    print(j.interpj('HCHO_M_SAPRC99', 500, 15, .5))



__all__ = ['arlpardump']
from PseudoNetCDF import PseudoNetCDFFile
import numpy as np
from datetime import datetime


class arlpardump(PseudoNetCDFFile):
    @classmethod
    def isMine(cls, path):
        try:
            arlpardump(path)
            return True
        except Exception:
            return False

    def __init__(self, path):
        self._path = path
        self._file = open(path, 'rb')
        self._read()
        mass = self._data['data']['f1']
        loc = self._data['data']['f4']
        meta = self._data['data']['f7']
        self.createDimension('particles', self.NPARTICLES)
        self.createDimension('pollutants', self.NPOLLUTANTS)
        startdstr = '{:02d}{:02d}{:02d}{:02d}{:02d}+0000'.format(
            self.YEAR, self.MONTH, self.DAY, self.HOUR, self.MINUTES)
        startd = datetime.strptime(startdstr, '%y%m%d%H%M%z')
        pvar = self.createVariable(
            'particle_mass', 'f', ('particles', 'pollutants'), values=mass)
        pvar.units = 'arbitrary'
        pvar.long_name = 'PARTICLE_MASS'
        var = self.createVariable(
            'latitude', 'f', ('particles',), values=loc[:, 0])
        var.units = 'degrees_north'
        var.long_name = 'latitude'
        var = self.createVariable(
            'longitude', 'f', ('particles',), values=loc[:, 1])
        var.units = 'degrees_east'
        var.long_name = 'latitude'
        var = self.createVariable(
            'height', 'f', ('particles',), values=loc[:, 2])
        var.units = 'meters'
        var.long_name = 'HEIGHT'
        var = self.createVariable(
            'sigma_u', 'f', ('particles',), values=loc[:, 3])
        var.units = 'sigma'
        var.long_name = 'SIGMA-U'
        var = self.createVariable(
            'sigma_v', 'f', ('particles',), values=loc[:, 4])
        var.units = 'sigma'
        var.long_name = 'SIGMA-V'
        var = self.createVariable(
            'sigma_x', 'f', ('particles',), values=loc[:, 5])
        var.units = 'sigma'
        var.long_name = 'SIGMA-X'
        var = self.createVariable(
            'age', 'i', ('particles',), values=meta[:, 0])
        var.units = 'minutes since {}'.format(
            startd.strftime('%Y-%m-%d %H:%M:%S%z'))
        var.long_name = 'AGE'
        var = self.createVariable(
            'distribution', 'i', ('particles',), values=meta[:, 1])
        var.units = '---'
        var.long_name = 'DISTRIBUTION'
        var = self.createVariable(
            'pollutant', 'i', ('particles',), values=meta[:, 2])
        var.units = '---'
        var.long_name = 'POLLUTANT'
        var = self.createVariable(
            'meteo_grid', 'i', ('particles',), values=meta[:, 3])
        var.units = '---'
        var.long_name = 'METEO-GRID'
        var = self.createVariable(
            'sort_index', 'i', ('particles',), values=meta[:, 4])
        var.units = '---'
        var.long_name = 'SORT-INDEX'

    def _read(self):
        """
        INT*4 Number of particles
        INT*4 Number of pollutants
        INT*4 Time of particle dump (YEAR, MONTH, DAY, HOUR, MINUTES)
        """
        rec1 = np.fromfile(self._path, dtype='>i,>7i,>i', count=1)
        self._file.seek(0, 0)
        hdr = rec1['f1'][0]
        self.NPARTICLES = hdr[0]
        self.NPOLLUTANTS = hdr[1]
        self.YEAR = hdr[2]
        self.MONTH = hdr[3]
        self.DAY = hdr[4]
        self.HOUR = hdr[5]
        self.MINUTES = hdr[6]
        assert(rec1['f0'] == rec1['f2'])
        datfmt = '>i,>({},)f,>i,>i,>6f,>i,>i,>5i,>i'.format(self.NPOLLUTANTS)
        self._fmt = np.dtype([('header', '>i,>7i,>i'),
                              ('data', datfmt, (self.NPARTICLES,))])
        data = np.fromfile(self._path, dtype=self._fmt)
        assert(data.shape[0] == 1)
        self._data = data[0]
        db = self._data['data']
        assert((db['f0'] == db['f2']).all())
        assert((db['f3'] == db['f5']).all())
        assert((db['f6'] == db['f8']).all())


if __name__ == '__main__':
    from PseudoNetCDF import pncopen
    f = pncopen('PARDUMP_143', format='arlpardump')



from unittest import TestSuite, findTestCases, TextTestRunner
from ..core import _files
from .. import sci_var
from .. import ArrayTransforms
from .. import camxfiles
from .. import net_balance
from .. import geoschemfiles
from .. import textfiles
from .. import icarttfiles
from .. import _getreader


test_suite = TestSuite()


def addTestCasesFromModule(module):
    test_suite.addTests(findTestCases(module))


# Core file tests
addTestCasesFromModule(_files)

# sci_var tests
addTestCasesFromModule(sci_var)

# ArrayTransforms ests
addTestCasesFromModule(ArrayTransforms)

# CAMx tests
addTestCasesFromModule(camxfiles.wind.Memmap)
addTestCasesFromModule(camxfiles.humidity.Memmap)
addTestCasesFromModule(camxfiles.temperature.Memmap)
addTestCasesFromModule(camxfiles.vertical_diffusivity.Memmap)
addTestCasesFromModule(camxfiles.one3d.Memmap)
addTestCasesFromModule(camxfiles.height_pressure.Memmap)
addTestCasesFromModule(camxfiles.landuse.Memmap)
addTestCasesFromModule(camxfiles.cloud_rain.Memmap)
addTestCasesFromModule(camxfiles.uamiv.Memmap)
addTestCasesFromModule(camxfiles.uamiv.Write)
addTestCasesFromModule(camxfiles.point_source.Memmap)
addTestCasesFromModule(camxfiles.lateral_boundary.Memmap)
addTestCasesFromModule(camxfiles.ipr.Memmap)
addTestCasesFromModule(camxfiles.irr.Memmap)
addTestCasesFromModule(camxfiles.FortranFileUtil)
addTestCasesFromModule(camxfiles.timetuple)

# Net balance test
addTestCasesFromModule(net_balance)

# GEOS-Chem tests
addTestCasesFromModule(geoschemfiles._bpch)
addTestCasesFromModule(geoschemfiles._newbpch)
addTestCasesFromModule(geoschemfiles._geos)

# Text file tests
addTestCasesFromModule(textfiles._delimited)

# ICARTT Files
addTestCasesFromModule(icarttfiles.ffi1001)

# Reader tests
addTestCasesFromModule(_getreader)


def test_all():
    TextTestRunner(verbosity=2).run(test_suite)


if __name__ == '__main__':
    test_all()



__all__ = ['wrf_base', 'wrf']
from ..core._files import PseudoNetCDFFile, netcdf


class wrf_base(PseudoNetCDFFile):
    @classmethod
    def isMine(self, path, *args, **kwds):
        return False

    def getTimes(self):
        cf = self.getCoordFile()
        return PseudoNetCDFFile.getTimes(cf)

    def getCoordFile(self):
        from ..conventions.ioapi import add_cf_from_wrfioapi
        cf = self.subsetVariables(list(self.getCoords()))
        add_cf_from_wrfioapi(cf)
        coordkeys = [k for k in cf.variables if k in cf.dimensions]
        coordkeys.extend(['longitude', 'latitude', 'time'])
        cf.setCoords(coordkeys)
        return cf

    def __init__(self, *args, **kwds):
        super().__init__(self, *args, **kwds)
        self._defCoords()

    def _defCoords(self):
        coordkeys = [
            key for key in self.variables
            if (
                key.startswith('XLAT') or key.startswith('XLONG') or
                key.startswith('Times')
            )
        ]
        self.setCoords(coordkeys)

    def plot(self, varkey, **kwds):
        var = self.variables[varkey]
        mykwds = kwds.copy()
        map_kw = kwds.pop('map_kw', None)
        mykwds['map_kw'] = False

        dimopts = [
            dk for dk in var.dimensions
            if len(self.dimensions[dk]) > 0
        ]
        plottype = '-'.join(dimopts[-2:][::-1])
        mykwds.setdefault('plottype', plottype)

        cf = self.getCoordFile()
        cf.copyVariable(var, key=varkey)
        ax = PseudoNetCDFFile.plot(cf, varkey, **kwds)
        maptypes = ('west_east-south_north', 'west_east_stag-south_north_stag')

        if plottype in maptypes and map_kw is not False:
            if map_kw is True or map_kw is None:
                map_kw = {}
            mymap_kw = map_kw.copy()
            mymap_kw.pop('coastlines', True)
            mymap_kw.pop('countries', True)
            mymap_kw.pop('states', False)
            mymap_kw.pop('counties', False)
            bmap = self.getMap(**mymap_kw)
            if map_kw.get('coastlines', True):
                bmap.drawcoastlines(ax=ax)
            if map_kw.get('countries', True):
                bmap.drawcountries(ax=ax)
            if map_kw.get('states', False):
                bmap.drawstates(ax=ax)
            if map_kw.get('counties', False):
                bmap.drawcounties(ax=ax)

        return ax

    def val2idx(self, *args, **kwds):
        cf = self.getCoordFile()
        return PseudoNetCDFFile.val2idx(cf, *args, **kwds)

    def xy2ll(self, *args, **kwds):
        cf = self.getCoordFile()
        return PseudoNetCDFFile.xy2ll(cf, *args, **kwds)

    def ll2xy(self, *args, **kwds):
        cf = self.getCoordFile()
        return PseudoNetCDFFile.ll2xy(cf, *args, **kwds)

    def ll2ij(self, *args, **kwds):
        cf = self.getCoordFile()
        return PseudoNetCDFFile.ll2ij(cf, *args, **kwds)

    def ij2ll(self, *args, **kwds):
        cf = self.getCoordFile()
        return PseudoNetCDFFile.ij2ll(cf, *args, **kwds)

    def getMap(self, maptype='basemap_auto', **kwds):
        mykwds = kwds.copy()
        cf = self.getCoordFile()
        if maptype.endswith('_auto'):
            x = cf.variables['x']
            y = cf.variables['y']
            xmax = x.max() - x.min()
            ymax = y.max() - y.min()
            lllon, lllat = cf.xy2ll(0, 0)
            urlon, urlat = cf.xy2ll(xmax, ymax)
            mykwds.setdefault('llcrnrlon', lllon)
            mykwds.setdefault('llcrnrlat', lllat)
            mykwds.setdefault('urcrnrlon', urlon)
            mykwds.setdefault('urcrnrlat', urlat)
            maptype = maptype[:-5]

        return PseudoNetCDFFile.getMap(cf, maptype=maptype, **mykwds)


class wrf(wrf_base, netcdf):
    def __init__(self, *args, **kwds):
        netcdf.__init__(self, *args, **kwds)
        self._defCoords()

    def _newlike(self):
        if self.get_dest() is not None:
            outf = wrf(**self.get_dest())
        elif isinstance(self, PseudoNetCDFFile):
            outt = wrf_base
            outf = outt.__new__(outt)
        else:
            outf = PseudoNetCDFFile()
        outf.set_varopt(**self.get_varopt())
        return outf

    @classmethod
    def from_ncf(cls, infile):
        outf = wrf_base()
        for pk in infile.ncattrs():
            pv = getattr(infile, pk)
            setattr(outf, pk, pv)

        for dk, dv in infile.dimensions.items():
            outf.copyDimension(dv, key=dk)

        for vk, vv in infile.variables.items():
            outf.copyVariable(vv, key=vk)

        return outf

    @property
    def _mode(self):
        return self.__dict__['_mode']

    def createVariable(self, *args, **kwds):
        return netcdf.createVariable(self, *args, **kwds)

    def createDimension(self, *args, **kwds):
        return netcdf.createDimension(self, *args, **kwds)

    @classmethod
    def isMine(cls, *args, **kwds):
        try:
            f = netcdf(*args, **kwds)
            dimkeys = ['Time', 'bottom_top', 'west_east', 'north_south']
            for dimkey in dimkeys:
                if dimkey not in f.dimensions:
                    return False
            else:
                return True
        except Exception:
            return False



from __future__ import print_function
__all__ = ['finst']
__doc__ = """
.. _Memmap
:mod:`Memmap` -- finst Memmap interface
============================================

.. module:: Memmap
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map for CAMx fine
              instantaneus/avg files.  See PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

from warnings import warn
# Distribution packages
import unittest
import struct

# Site-Packages
from numpy import zeros, memmap, dtype, fromfile

# This Package modules
from PseudoNetCDF.camxfiles.units import get_uamiv_units
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariable
from PseudoNetCDF.sci_var import PseudoNetCDFVariables
from PseudoNetCDF.ArrayTransforms import ConvertCAMxTime

# for use in identifying uncaught nan
listnan = struct.unpack('>f', b'\xff\xc0\x00\x00')[0]
checkarray = zeros((1,), 'f')
checkarray[0] = listnan
array_nan = checkarray[0]


class finst(PseudoNetCDFFile):
    """
    finst provides a PseudoNetCDF interface for CAMx
    finst files.  Where possible, the inteface follows
    IOAPI conventions (see www.baronams.com).

    ex:
        >>> finst_path = 'camx_finst.bin'
        >>> finstfile = finst(finst_path)
        >>> finstfile.variables.keys()
        ['TFLAG', 'O3', 'NO', 'NO2', ...]
        >>> v = finstfile.variables['V']
        >>> tflag = finstfile.variables['TFLAG']
        >>> tflag.dimensions
        ('TSTEP', 'VAR', 'DATE-TIME')
        >>> tflag[0,0,:]
        array([2005185,       0])
        >>> tflag[-1,0,:]
        array([2005185,  240000])
        >>> v.dimensions
        ('TSTEP', 'LAY', 'ROW', 'COL')
        >>> v.shape
        (25, 28, 65, 83)
        >>> finstfile.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
    """

    __messagefmt = '>240S'
    __nestspcfmt = dtype(dict(names=['nnest', 'nspec'], formats=['>i', '>i']))
    __nestparmsfmt = dtype(dict(names=['SPAD', 'ibeg', 'jbeg', 'iend', 'jend',
                                       'mesh', 'ione', 'nx', 'ny', 'nz',
                                       'iparent', 'ilevel', 'EPAD'],
                                formats=['>i'] * 13))
    __time_hdr_fmt = dtype(dict(names=['time', 'date'], formats=['>f', '>i']))
    __spc_fmt = dtype(">10S")
    __ione = 1
    __idum = 0
    __rdum = 0.
    __buffersize = 4

    def __init__(self, rf, mode='r'):
        """
        Initialization included reading the header and learning
        about the format.

        see __readheader and __gettimestep() for more info
        """
        self.__rffile = rf

        # Establish dimensions
        self.createDimension('DATE-TIME', 2)

        self.__readheader(mode)

        # Add IOAPI metavariables
        self.NLAYS = len(self.dimensions['LAY'])
        self.NROWS = len(self.dimensions['ROW'])
        self.NCOLS = len(self.dimensions['COL'])
        self.NVARS = len(self.dimensions['VAR'])
        self.NSTEPS = len(self.dimensions['TSTEP'])
        varlist = "".join([i.ljust(16) for i in self.__var_names__])
        setattr(self, 'VAR-LIST', varlist)
        self.GDTYP = 2

        # Create variables
        self.variables = PseudoNetCDFVariables(
            self.__variables, self.__var_names__ + ['TFLAG'])

        # Initialize time maps
        date = self.__memmap__['DATE']['DATE']
        time = self.__memmap__['DATE']['TIME']
        self.variables['TFLAG'] = ConvertCAMxTime(date, time, self.NVARS)
        self.SDATE, self.STIME = self.variables['TFLAG'][0, 0, :]

    def __checkfilelen(self):
        f = open(self.__rffile, 'rb')
        f.seek(0, 2)
        flen = f.tell()
        f.close()
        return flen

    def __readheader(self, mode):
        f = open(self.__rffile)
        f.seek(92)
        self.NNEST, self.NSPEC = fromfile(f, '>i', 2)
        f.seek(108)

        self.SPECIES = fromfile(f, self.__spc_fmt, self.NSPEC)

        offset = f.tell() + 4
        f.close()
        del f
        self.__var_names__ = [
            i.decode().strip() for i in self.SPECIES.tolist()
        ]

        self.NEST_PARMS = memmap(
            self.__rffile, dtype=self.__nestparmsfmt, mode=mode,
            offset=offset, shape=self.NNEST)

        offset += self.__nestparmsfmt.itemsize * self.NNEST

        date_fmt = dtype(
            dict(names=['SPAD', 'TIME', 'DATE', 'EPAD'],
                 formats=['>i', '>f', '>i', '>i']))
        spc_1_lay_fmt = []
        for i in range(self.NNEST):
            nny = self.NEST_PARMS[i]['ny']
            nnx = self.NEST_PARMS[i]['nx']
            nnfmts = ['>i', '(%d,%d)>f' % (nny, nnx), '>i']
            spc_1_lay_fmt.append(dtype(dict(names=['SPAD', 'DATA', 'EPAD'],
                                            formats=nnfmts)))

        grid_fmt = []
        for i in range(self.NNEST):
            gfmts = [dtype((spc_1_lay_fmt[i],
                            (int(self.NEST_PARMS[i]['nz']), )))] * self.NSPEC
            grid_fmt.append(dtype(dict(names=self.__var_names__,
                                       formats=gfmts)))

        mmnames = ['DATE'] + ['grid%d' % i for i in range(self.NNEST)]
        mmdt = dtype(dict(names=mmnames, formats=[date_fmt] + grid_fmt))
        self.__memmap__ = memmap(self.__rffile, mode=mode, offset=offset,
                                 dtype=mmdt)

        ntimes = self.__memmap__.shape[0]
        if int(ntimes) != ntimes:
            raise ValueError("Not an even number of times")

        self.createDimension('TSTEP', ntimes)
        self.createDimension('VAR', self.NSPEC)
        self.chooseGrid(0)

    def chooseGrid(self, ngrid):
        self.createDimension('LAY', self.NEST_PARMS[ngrid]['nz'])
        self.createDimension('COL', self.NEST_PARMS[ngrid]['nx'])
        self.createDimension('ROW', self.NEST_PARMS[ngrid]['ny'])
        self.CURRENT_GRID = 'grid%d' % ngrid

    def __variables(self, k):
        dimensions = ('TSTEP', 'LAY', 'ROW', 'COL')

        outvals = self.__memmap__[self.CURRENT_GRID][k]['DATA'][:, :, :, :]
        unit = get_uamiv_units('INSTANT   ', k)
        return PseudoNetCDFVariable(self, k, 'f', dimensions,
                                    values=outvals, units=unit)


class TestMemmap(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testFinst(self):
        import PseudoNetCDF.testcase
        finstfile = finst(PseudoNetCDF.testcase.camxfiles_paths['finst'])
        vars = [i for i in finstfile.variables.keys() if i != 'TFLAG']
        for var in vars:
            v = finstfile.variables[var]
            warn("Test case is not fully implemented.  " +
                 "Review values for 'reasonability.'")
            datetime = finstfile.variables['TFLAG']
            minv = v.min(1).min(1).min(1)
            meanv = v.mean(1).mean(1).mean(1)
            maxv = v.max(1).max(1).max(1)
            for i, (d, t) in enumerate(datetime[:, 0, :]):
                print(var, d, t, minv[i], meanv[i], maxv[i])


if __name__ == '__main__':
    unittest.main()



from __future__ import print_function  # , unicode_literals
from PseudoNetCDF.sci_var import PseudoNetCDFFile
import numpy as np
from warnings import warn
from PseudoNetCDF._getwriter import registerwriter
import unittest

_coordkeys = ("time time_bounds TFLAG ETFLAG latitude latitude_bounds " +
              "longitude longitude_bounds lat lat_bnds lon lon_bnds " +
              "etam_pressure etai_pressure layer_bounds layer47 " +
              "layer").split()


class csv(PseudoNetCDFFile):
    def __init__(self, path, coordkeys=_coordkeys, delimiter=',',
                 names=True, backend=None, defaultcoord='record', **kwds):
        """
        path : str
            place to find csv file
        coordkeys : iterable of strings
            use these keys as dimensions and coordinate variables
        delimiter : str
            use this as delimiter (default = ',') renamed as sep for pandas
        names : iterable of strings or True
            with pandas backend and names==True: header='infer'
            otherwise, passed directly as keyword
        backend : str or None
            'numpy' numpy.recfromtxt or 'pandas' pandas.read_csv; defaults to
            pandas if available
        defaultcoord : str
            if no coordkeys are found, use this str to create a arbitrary
            coordinate based on each record
        kwds : mappable
            corresponds to numpy.recfromtxt or pandas.read_csv keywords

        * Note: currently only works when all coordinate variables are 1-d
        """
        if backend != 'numpy':
            try:
                import pandas
                if backend is None:
                    backend = 'pandas'
            except Exception:
                if backend == 'pandas':
                    raise ValueError(
                        'pandas library not available, try another backend'
                    )

        if backend == 'numpy':
            npkwds = kwds.copy()
            npkwds['names'] = names
            npkwds['delimiter'] = delimiter
            data = np.recfromtxt(path, **npkwds)
        elif backend == 'pandas':
            pdkwds = kwds.copy()
            if names is True:
                pdkwds.setdefault('header', 'infer')
            else:
                pdkwds['names'] = names
            pdkwds.setdefault('sep', delimiter)
            odata = pandas.read_csv(path, **pdkwds)
            # pandas leaves whitespace in names, which is not good for
            # netcdf-like names and probably not intended.
            # odata.rename(columns=lambda x: x.strip(), inplace=True)
            # relying on user to supply skipinitialspace=True
            data = odata.to_records(index=False)
        else:
            raise ValueError(
                "backend options are 'numpy' or 'pandas': got '%s'" % backend
            )

        varkeys = [vk for vk in data.dtype.names if vk not in coordkeys]
        dimkeys = tuple([dk for dk in coordkeys if dk in data.dtype.names])
        dimvars = {}
        if len(dimkeys) == 0:
            dimkeys = (defaultcoord,)
            dv = self.createDimension(defaultcoord, data.shape[0])
            dvar = self.createVariable(defaultcoord, 'd', (defaultcoord,))
            dvar[:] = dimvars[defaultcoord] = np.arange(len(dv))
        else:
            for dk in dimkeys:
                dimvars[dk] = data[dk]
                dv = np.unique(data[dk])
                dv.sort()
                self.createDimension(dk, len(dv))
                mydtype = dv.dtype.char
                if mydtype == 'S':
                    mydtype = dv.dtype
                dvar = self.createVariable(dk, mydtype, (dk,))
                dvar[:] = dv

        for vk in varkeys:
            vv = data[vk]
            if vv.dtype.char != 'S':
                var = self.createVariable(
                    vk, vv.dtype.char, dimkeys, fill_value=-999)
                var[:] = -999
            else:
                var = self.createVariable(vk, vv.dtype.char, dimkeys)

        bigidx = []
        bigidx = dict([(dk, (dimvars[dk][:, None] == self.variables[dk]
                             [None, :]).argmax(1)) for dk in dimkeys])
        for vk in varkeys:
            vv = data[vk]
            ov = self.variables[vk]
            myidx = tuple([bigidx[dk] for dk in ov.dimensions])
            ov[myidx] = vv


def ncf2csv(ifile, outpath, delimiter=',', coordkeys=_coordkeys):
    header = [k for k, v in ifile.variables.items(
    ) if k not in coordkeys and v.size > 1 and k not in ifile.dimensions]
    dims = set([ifile.variables[k].dimensions for k in header])
    if len(dims) > 1:
        if hasattr(outpath, 'write'):
            warn('Multiple csv outputs will be separated by ### because ' +
                 'not all output variables have the same dimensions')
        else:
            warn('Making multiple csv outputs because not all output ' +
                 'variables have the same dimensions')
    dimsets = {}
    for di, dim in enumerate(dims):
        if len(dims) > 1:
            if not hasattr(outpath, 'write'):
                outfile = open(outpath + str(di), 'wt')
            else:
                print('###', file=outfile)
        else:
            if not hasattr(outpath, 'write'):
                outfile = open(outpath, mode='wt')
            else:
                outfile = outpath

        dimheader = [k for k in dim if k in ifile.variables]
        header = dimsets[dim] = [
            k for k, v in ifile.variables.items() if v.dimensions == dim]
        dimvars = [ifile.variables[k] for k in dimheader]
        vars = [ifile.variables[k] for k in header]
        outtext = delimiter.join(dimheader + header)
        print(outtext, file=outfile)
        for idx in np.ndindex(ifile.variables[header[-1]].shape):
            outvals = []
            for dk, dv in zip(dimheader, dimvars):
                dv = ifile.variables[dk]
                didx = tuple([iidx for i, iidx in enumerate(
                    idx) if dim[i] in dv.dimensions])
                outvals.append(repr(dv[didx]))
            for vk, vv in zip(header, vars):
                outvals.append(repr(vv[idx]))
            outtext = delimiter.join(outvals)
            print(outtext, file=outfile)


registerwriter('csv', ncf2csv)


class TestCsv(unittest.TestCase):
    def setUp(self):
        from PseudoNetCDF import PseudoNetCDFFile
        self.checkval = """time,layer,latitude,longitude,test
0.0,0.0,0.0,0.0,0.0
0.0,0.0,0.0,1.0,1.0
0.0,0.0,0.0,2.0,2.0
0.0,0.0,0.0,3.0,3.0
0.0,0.0,0.0,4.0,4.0
0.0,0.0,1.0,0.0,5.0
0.0,0.0,1.0,1.0,6.0
0.0,0.0,1.0,2.0,7.0
0.0,0.0,1.0,3.0,8.0
0.0,0.0,1.0,4.0,9.0
0.0,0.0,2.0,0.0,10.0
0.0,0.0,2.0,1.0,11.0
0.0,0.0,2.0,2.0,12.0
0.0,0.0,2.0,3.0,13.0
0.0,0.0,2.0,4.0,14.0
0.0,0.0,3.0,0.0,15.0
0.0,0.0,3.0,1.0,16.0
0.0,0.0,3.0,2.0,17.0
0.0,0.0,3.0,3.0,18.0
0.0,0.0,3.0,4.0,19.0
0.0,1.0,0.0,0.0,20.0
0.0,1.0,0.0,1.0,21.0
0.0,1.0,0.0,2.0,22.0
0.0,1.0,0.0,3.0,23.0
0.0,1.0,0.0,4.0,24.0
0.0,1.0,1.0,0.0,25.0
0.0,1.0,1.0,1.0,26.0
0.0,1.0,1.0,2.0,27.0
0.0,1.0,1.0,3.0,28.0
0.0,1.0,1.0,4.0,29.0
0.0,1.0,2.0,0.0,30.0
0.0,1.0,2.0,1.0,31.0
0.0,1.0,2.0,2.0,32.0
0.0,1.0,2.0,3.0,33.0
0.0,1.0,2.0,4.0,34.0
0.0,1.0,3.0,0.0,35.0
0.0,1.0,3.0,1.0,36.0
0.0,1.0,3.0,2.0,37.0
0.0,1.0,3.0,3.0,38.0
0.0,1.0,3.0,4.0,39.0
0.0,2.0,0.0,0.0,40.0
0.0,2.0,0.0,1.0,41.0
0.0,2.0,0.0,2.0,42.0
0.0,2.0,0.0,3.0,43.0
0.0,2.0,0.0,4.0,44.0
0.0,2.0,1.0,0.0,45.0
0.0,2.0,1.0,1.0,46.0
0.0,2.0,1.0,2.0,47.0
0.0,2.0,1.0,3.0,48.0
0.0,2.0,1.0,4.0,49.0
0.0,2.0,2.0,0.0,50.0
0.0,2.0,2.0,1.0,51.0
0.0,2.0,2.0,2.0,52.0
0.0,2.0,2.0,3.0,53.0
0.0,2.0,2.0,4.0,54.0
0.0,2.0,3.0,0.0,55.0
0.0,2.0,3.0,1.0,56.0
0.0,2.0,3.0,2.0,57.0
0.0,2.0,3.0,3.0,58.0
0.0,2.0,3.0,4.0,59.0
1.0,0.0,0.0,0.0,60.0
1.0,0.0,0.0,1.0,61.0
1.0,0.0,0.0,2.0,62.0
1.0,0.0,0.0,3.0,63.0
1.0,0.0,0.0,4.0,64.0
1.0,0.0,1.0,0.0,65.0
1.0,0.0,1.0,1.0,66.0
1.0,0.0,1.0,2.0,67.0
1.0,0.0,1.0,3.0,68.0
1.0,0.0,1.0,4.0,69.0
1.0,0.0,2.0,0.0,70.0
1.0,0.0,2.0,1.0,71.0
1.0,0.0,2.0,2.0,72.0
1.0,0.0,2.0,3.0,73.0
1.0,0.0,2.0,4.0,74.0
1.0,0.0,3.0,0.0,75.0
1.0,0.0,3.0,1.0,76.0
1.0,0.0,3.0,2.0,77.0
1.0,0.0,3.0,3.0,78.0
1.0,0.0,3.0,4.0,79.0
1.0,1.0,0.0,0.0,80.0
1.0,1.0,0.0,1.0,81.0
1.0,1.0,0.0,2.0,82.0
1.0,1.0,0.0,3.0,83.0
1.0,1.0,0.0,4.0,84.0
1.0,1.0,1.0,0.0,85.0
1.0,1.0,1.0,1.0,86.0
1.0,1.0,1.0,2.0,87.0
1.0,1.0,1.0,3.0,88.0
1.0,1.0,1.0,4.0,89.0
1.0,1.0,2.0,0.0,90.0
1.0,1.0,2.0,1.0,91.0
1.0,1.0,2.0,2.0,92.0
1.0,1.0,2.0,3.0,93.0
1.0,1.0,2.0,4.0,94.0
1.0,1.0,3.0,0.0,95.0
1.0,1.0,3.0,1.0,96.0
1.0,1.0,3.0,2.0,97.0
1.0,1.0,3.0,3.0,98.0
1.0,1.0,3.0,4.0,99.0
1.0,2.0,0.0,0.0,100.0
1.0,2.0,0.0,1.0,101.0
1.0,2.0,0.0,2.0,102.0
1.0,2.0,0.0,3.0,103.0
1.0,2.0,0.0,4.0,104.0
1.0,2.0,1.0,0.0,105.0
1.0,2.0,1.0,1.0,106.0
1.0,2.0,1.0,2.0,107.0
1.0,2.0,1.0,3.0,108.0
1.0,2.0,1.0,4.0,109.0
1.0,2.0,2.0,0.0,110.0
1.0,2.0,2.0,1.0,111.0
1.0,2.0,2.0,2.0,112.0
1.0,2.0,2.0,3.0,113.0
1.0,2.0,2.0,4.0,114.0
1.0,2.0,3.0,0.0,115.0
1.0,2.0,3.0,1.0,116.0
1.0,2.0,3.0,2.0,117.0
1.0,2.0,3.0,3.0,118.0
1.0,2.0,3.0,4.0,119.0
"""
        testfile = self.testfile = PseudoNetCDFFile()
        testfile.createDimension('time', 2)
        testfile.createDimension('layer', 3)
        testfile.createDimension('latitude', 4)
        testfile.createDimension('longitude', 5)
        for dk, dv in testfile.dimensions.items():
            var = testfile.createVariable(dk, 'f', (dk,))
            var[:] = np.arange(len(dv), dtype='f')
        var = testfile.createVariable(
            'test', 'f', ('time', 'layer', 'latitude', 'longitude'))
        var[:] = np.arange(2 * 3 * 4 * 5).reshape(2, 3, 4, 5)

    def testNCF2CSV(self):
        from PseudoNetCDF.pncgen import pncgen
        from io import BytesIO
        import tempfile
        out = BytesIO()
        out = tempfile.TemporaryFile(mode='w+t')
        pncgen(self.testfile, out, inmode='r',
               outmode='w', format='csv', verbose=0)
        out.seek(0, 0)
        outval = out.read()
        assert(outval == self.checkval)



__all__ = ['aerosol_names']
__doc__ = """
mech_aerosol_options keys:
    camx6_3_chemparam_2_cf
    camx6_3_chemparam_6_cmu
    camx6_3_chemparam_2_cf_vbs
    camx6_3_chemparam_6_none
    camx6_3_chemparam_2_none
    camx6_3_chemparam_4_none
    camx6_3_chemparam_6_cf_hg
    camx6_3_chemparam_3_none
    camx6_3_chemparam_6_cf
    camx6_3_chemparam_6_cf_vbs
    camx6_3_chemparam_4_cf
    camx6_3_chemparam_4_none_srfmod
    camx6_3_chemparam_5_cf
    camx6_3_chemparam_5_none
    camx6_3_chemparam_3_cf
    camx6_3_chemparam_6_inert
"""
mech_aerosol_options = {
    "camx6_3_chemparam_2_cf": ('PNO3', 'PSO4', 'PNH4', 'POA', 'SOA1', 'SOA2',
                               'SOA3', 'SOA4', 'SOA5', 'SOA6', 'SOA7', 'SOAH',
                               'SOPA', 'SOPB', 'PEC', 'FPRM', 'FCRS', 'CPRM',
                               'CCRS', 'NA', 'PCL', 'PH2O'),
    "camx6_3_chemparam_2_cf_vbs": ('PNO3', 'PSO4', 'PNH4', 'PAS0', 'PAS1',
                                   'PAS2', 'PAS3', 'PAS4', 'PBS0', 'PBS1',
                                   'PBS2', 'PBS3', 'PBS4', 'PAP0', 'PAP1',
                                   'PAP2', 'PAP3', 'PAP4', 'PCP0', 'PCP1',
                                   'PCP2', 'PCP3', 'PCP4', 'PFP0', 'PFP1',
                                   'PFP2', 'PFP3', 'PFP4', 'PEC', 'FPRM',
                                   'FCRS', 'CPRM', 'CCRS', 'NA', 'PCL',
                                   'PH2O'),
    "camx6_3_chemparam_2_none": (),
    "camx6_3_chemparam_3_cf": ('PNO3', 'PSO4', 'PNH4', 'POA', 'SOA1', 'SOA2',
                               'SOA3', 'SOA4', 'SOA5', 'SOA6', 'SOA7', 'SOAH',
                               'SOPA', 'SOPB', 'PEC', 'FPRM', 'FCRS', 'CPRM',
                               'CCRS', 'NA', 'PCL', 'PH2O'),
    "camx6_3_chemparam_3_none": (),
    "camx6_3_chemparam_4_cf": ('PNO3', 'PSO4', 'PNH4', 'POA', 'SOA1', 'SOA2',
                               'SOA3', 'SOA4', 'SOA5', 'SOA6', 'SOA7', 'SOAH',
                               'SOPA', 'SOPB', 'PEC', 'FPRM', 'FCRS', 'CPRM',
                               'CCRS', 'NA', 'PCL', 'PH2O'),
    "camx6_3_chemparam_4_none": (),
    "camx6_3_chemparam_4_none_srfmod": (),
    "camx6_3_chemparam_5_cf": ('PNO3', 'PSO4', 'PNH4', 'POA', 'SOA1', 'SOA2',
                               'SOA3', 'SOA4', 'SOA5', 'SOA6', 'SOA7', 'SOAH',
                               'SOPA', 'SOPB', 'PEC', 'FPRM', 'FCRS', 'CPRM',
                               'CCRS', 'NA', 'PCL', 'PH2O'),
    "camx6_3_chemparam_5_none": (),
    "camx6_3_chemparam_6_cf": ('PNO3', 'PSO4', 'PNH4', 'POA', 'SOA1', 'SOA2',
                               'SOA3', 'SOA4', 'SOA5', 'SOA6', 'SOA7', 'SOAH',
                               'SOPA', 'SOPB', 'PEC', 'FPRM', 'FCRS', 'CPRM',
                               'CCRS', 'NA', 'PCL', 'PH2O'),
    "camx6_3_chemparam_6_cf_hg": ('PNO3', 'PSO4', 'PNH4', 'POA', 'SOA1',
                                  'SOA2', 'SOA3', 'SOA4', 'SOA5', 'SOA6',
                                  'SOA7', 'SOAH', 'SOPA', 'SOPB', 'PEC',
                                  'FPRM', 'FCRS', 'CPRM', 'CCRS', 'NA', 'PCL',
                                  'PH2O', 'HGP', 'HGIIP', 'HGIIPC'),
    "camx6_3_chemparam_6_cf_vbs": ('PNO3', 'PSO4', 'PNH4', 'PAS0', 'PAS1',
                                   'PAS2', 'PAS3', 'PAS4', 'PBS0', 'PBS1',
                                   'PBS2', 'PBS3', 'PBS4', 'PAP0', 'PAP1',
                                   'PAP2', 'PAP3', 'PAP4', 'PCP0', 'PCP1',
                                   'PCP2', 'PCP3', 'PCP4', 'PFP0', 'PFP1',
                                   'PFP2', 'PFP3', 'PFP4', 'PEC', 'FPRM',
                                   'FCRS', 'CPRM', 'CCRS', 'NA', 'PCL',
                                   'PH2O'),
    "camx6_3_chemparam_6_cmu": ('PNO3', 'PSO4', 'PNH4', 'POA', 'SOA1', 'SOA2',
                                'SOA3', 'SOA4', 'SOA5', 'SOA6', 'SOA7', 'SOAH',
                                'SOPA', 'SOPB', 'PEC', 'CRST', 'NA', 'PCL',
                                'PH2O'),
    "camx6_3_chemparam_6_inert": ('FCRS', 'CCRS'),
    "camx6_3_chemparam_6_none": (),
}
aerosol_names = ('PEC', 'PNH4', 'PCP0', 'SOA1', 'PBS0', 'SOA5', 'FCRS', 'PH2O',
                 'SOPB', 'PBS3', 'PFP4', 'PCL', 'SOA3', 'HGIIPC', 'SOPA',
                 'PAS0', 'PFP1', 'PAP4', 'CPRM', 'SOA7', 'SOA6', 'NA', 'PAP1',
                 'PBS1', 'HGIIP', 'PCP4', 'CRST', 'SOAH', 'FPRM', 'SOA2',
                 'PAS4', 'PAS2', 'PAS3', 'PAS1', 'POA', 'PFP0', 'SOA4', 'HGP',
                 'PAP2', 'PCP3', 'CCRS', 'PCP1', 'PFP3', 'PBS2', 'PSO4',
                 'PNO3', 'PFP2', 'PCP2', 'PBS4', 'PAP3', 'PAP0')



from PseudoNetCDF.pncwarn import warn
from ..core._files import PseudoNetCDFFile, netcdf
from collections import OrderedDict
from PseudoNetCDF._getwriter import registerwriter

import numpy as np
import datetime
today = datetime.datetime.today()
# Assuming some fo the most common
# options
_i = _ioapi_defaults = OrderedDict()
_i['IOAPI_VERSION'] = "N/A".ljust(80)
_i['EXEC_ID'] = "????????????????".ljust(80)
_i['FTYPE'] = 1
_i['CDATE'] = int(today.strftime('%Y%j'))
_i['CTIME'] = int(today.strftime('%H%M%S'))
_i['WDATE'] = int(today.strftime('%Y%j'))
_i['WTIME'] = int(today.strftime('%H%M%S'))
_i['NTHIK'] = 1
_i['GDTYP'] = 1
_i['P_ALP'] = 33.
_i['P_BET'] = 45.
_i['P_GAM'] = -97.
_i['XCENT'] = -97.
_i['YCENT'] = 40.
_i['VGTYP'] = 2
_i['VGTOP'] = np.float32(5000)
_i['GDNAM'] = "UNKNOWN         "
_i['UPNAM'] = "MAKEIOAPI       "
_i['FILEDESC'] = "".ljust(80)
_i['HISTORY'] = ""


def ioapi_sort_meta(infile):
    mydimensions = infile.dimensions.copy()
    outvars = getattr(infile, 'VAR-LIST', '').split()
    allvars = outvars + \
        [k for k in list(infile.variables)
         if k not in outvars and k != 'TFLAG']
    infile.NVARS = len(outvars)
    infile.dimensions = OrderedDict()
    for dk in 'TSTEP DATE-TIME LAY VAR ROW COL'.split():
        dv = mydimensions[dk]
        if dk == 'VAR':
            dl = infile.NVARS
        else:
            dl = len(dv)
        ndv = infile.createDimension(dk, dl)
        ndv.setunlimited(dv.isunlimited())
    myvariables = infile.variables
    infile.variables = OrderedDict()
    for vk in ['TFLAG'] + allvars:
        infile.variables[vk] = myvariables[vk]


class ioapi_base(PseudoNetCDFFile):
    @classmethod
    def isMine(self, path, *args, **kwds):
        return False

    def __init__(self, *args, **kwds):
        super().__init__(self, *args, **kwds)
        self.setCoords(['TFLAG'])

    def __setattr__(self, k, v):
        """
        IOAPI has expected data types, which largely conform to the NetCDF3
        classic data model. For example, integers are int32 and floats are
        float64. However, VGTOP and VGLVLS are exceptions where the data type
        is expected to be float32. Further, any character variables will have
        one of three lengths: 16, 80, 80 * 60.
        """
        if k == 'VGTOP':
            v = np.float32(v)
        elif k == 'VGLVLS':
            v = np.asarray(v, dtype='f')
        elif k in ('FILEDESC', 'HISTORY'):
            if len(v) > 4800:
                warn(f'Truncating {k} 4800')
                v = v[:4800]
            v = v.ljust(4800)
        elif k in ('GDNAM', 'UPNAM'):
            if len(v) > 16:
                warn(f'Truncating {k} to 16')
                v = v[:16]
            v = v.ljust(16)
        elif k in ('EXEC_ID',):
            if len(v) > 80:
                warn(f'Truncating {k} to 80')
                v = v[:80]
            v = v.ljust(80)

        super().__setattr__(k, v)

    def audit_meta(self, fail='warn'):
        """
        Audit the IOAPI metadata. Checks existence of properties, dimensions
        internal consistency of properties and dimensions, and character
        lengths

        Arguments
        ---------
        fail : bool
           If fail, 'warn' or 'error' or 'ignore'

        Returns
        -------
        passing, audit, var_audit:
            audit stats and audit checklist dictionaries

        """
        audit = {}
        dimlens = {}
        for dk, dv in self.dimensions.items():
            dimlens[dk] = len(dv)

        for dk in ['TSTEP', 'DATE-TIME', 'PERIM', 'LAY', 'VAR', 'ROW', 'COL']:
            audit[f'has_{dk}'] = dk in dimlens
            if dk in ('TSTEP', 'DATE-TIME', 'PERIM'):
                continue
            audit[f'has_N{dk}S'] = hasattr(self, f'N{dk}S')
            if audit[f'has_N{dk}S']:
                audit[dk] = getattr(self, f'N{dk}S', 0) == dimlens[dk]

        if audit['has_ROW'] and audit['has_COL'] and not audit['has_PERIM']:
            del audit['has_PERIM']
            hdims = ('ROW', 'COL')
        elif audit['has_PERIM'] and not (audit['has_COL'] or audit['has_ROW']):
            del audit['has_ROW']
            del audit['has_COL']
            hdims = ('PERIM',)

        varliststr = getattr(self, 'VAR-LIST')
        audit['VAR-LIST-LEN'] = (self.NVARS * 16) == len(varliststr)
        checkvarliststr = self.getVarlist(update=False, retval='str')
        checkvarlist = self.getVarlist(update=False, retval='list')
        asisvarlist = self.getVarlist(update=False, prune=False, retval='list')
        for varkey in asisvarlist:
            audit[f'has_{varkey}'] = varkey in self.variables

        audit['VAR-LIST'] = varliststr == checkvarliststr
        var_audits = {}
        proplens = {
            'units': 16,
            'long_name': 16,
            'var_desc': 80,
        }
        for vk, var in self.variables.items():
            var_audit = var_audits[vk] = {}
            vardims = var.dimensions
            varshape = var.shape
            if vk == 'TFLAG':
                var_audit['right_dims'] = (
                    vardims == ('TSTEP', 'VAR', 'DATE-TIME')
                )
            else:
                var_audit['right_dims'] = (
                    vardims == ('TSTEP', 'LAY') + hdims
                )
            var_audit['ndims'] = len(varshape) == len(vardims)
            for dk, vn in zip(vardims, varshape):
                var_audit[f'shape_{dk}'] = vn == dimlens[dk]

            for pk, pcl in proplens.items():
                var_audit[f'has_{pk}'] = hasattr(var, pk)
                var_audit[f'len_{pk}'] = len(getattr(var, pk, '')) == pcl

            var_audit['SUMMARY'] = all(var_audit.values())
            if vk in checkvarlist:
                var_audit['INCLUDED'] = True
                audit[f'var_{vk}'] = var_audit['SUMMARY']
            else:
                var_audit['INCLUDED'] = False

        for pk, pcv in _ioapi_defaults.items():
            hasp = hasattr(self, pk)
            audit[f'has_{pk}'] = hasp
            if hasp:
                typep = isinstance(getattr(self, pk), type(pcv))
            else:
                typep = False
            audit[f'type_{pk}'] = typep

        proplens = {
            'FILEDESC': 4800,
            'HISTORY': 4800
        }
        for pk, pcl in proplens.items():
            audit[f'has_{pk}'] = hasattr(self, pk)
            audit[f'len_{pk}'] = len(getattr(self, pk, '')) <= pcl

        hastflag = 'TFLAG' in self.variables
        if hastflag:
            t0 = self.variables['TFLAG'][0, 0]
            audit['SDATE_TFLAG'] = (t0[0] == self.SDATE) or self.SDATE == -635
            audit['STIME_TFLAG'] = t0[1] == self.STIME

        audit['has_TFLAG'] = hastflag
        passing = audit['SUMMARY'] = all(list(audit.values()))
        if not passing:
            if fail == 'ignore':
                pass
            elif fail == 'warn':
                warn('Failed meta data audit review settings.')
            else:
                failed = str({k: v for k, v in audit.items() if not v})
                var_failed = str({
                    vk: {k: v for k, v in vv.items() if not v}
                    for vk, vv in var_audits.items()
                })
                raise ValueError(f"""Failed audit.
File failures: {failed}
Varable failures: {var_failed}
""")
        return passing, audit, var_audits

    def _updatetime(self, write=True, create=False):
        from datetime import datetime
        t = datetime.now()
        try:
            if create:
                self.CDATE = int(t.strftime('%Y%j'))
                self.CTIME = int(t.strftime('%H%M%S'))
            if write:
                self.WDATE = int(t.strftime('%Y%j'))
                self.WTIME = int(t.strftime('%H%M%S'))
        except Exception as e:
            warn('Time could not be updated; ' + str(e))

    def setncatts(self, attdict):
        """
        Wrapper on PseudoNetCDF.setncatts that updates WDATE, and WTIME

        See also
        --------
        see PseudoNetCDFFile.setncatts
        """
        PseudoNetCDFFile.setncatts(self, attdict)
        self._updatetime()

    def createVariable(self, name, type='f', dimensions=None,
                       fill_value=None, **properties):
        """
        Wrapper on PseudoNetCDF.createVariable that updates VAR-LIST,
        NVARS, VAR, and TFLAG. Also adds long_name, var_desc, and units
        if not already in properties. long_name and var_desc default to
        name, while units defaults to unknown. These properties will
        also be adjusted to expected lengths (16, 80, 16).

        See also
        --------
        see PseudoNetCDFFile.createVariable
        """
        from copy import copy

        ftype = getattr(self, 'FTYPE', 1)
        hasperim = 'PERIM' in self.dimensions
        if dimensions is None:
            dimensions = ('TSTEP', 'LAY', 'ROW', 'COL')
            if (ftype == 2) and hasperim:
                dimensions = ('TSTEP', 'LAY', 'PERIM')

        properties = copy(properties)
        properties.setdefault('long_name', name.ljust(16))
        properties.setdefault('var_desc', name.ljust(80))
        properties.setdefault('units', 'unknown'.ljust(16))

        for pk, pv in list(properties.items()):
            if pk in ('long_name', 'units'):
                properties[pk] = pv.ljust(16)
            elif pk in ('var_desc',):
                properties[pk] = pv.ljust(80)

        if name == 'TFLAG':
            fill_value = None

        out = PseudoNetCDFFile.createVariable(
            self, name=name, type=type, dimensions=dimensions,
            fill_value=fill_value, **properties)

        self._add2Varlist([name])

        return out

    def copy(self, props=True, dimensions=True, variables=True, data=True):
        out = PseudoNetCDFFile.copy(
            self, props=props, dimensions=dimensions, variables=False,
            data=False
        )
        if variables:
            for vk, vv in self.variables.items():
                if not vk.endswith('TFLAG'):
                    PseudoNetCDFFile.copyVariable(
                        out, vv, key=vk, withdata=data
                    )

        if props and dimensions:
            out.updatetflag()

        return out

    def copyVariable(self, var, key=None, dtype=None, dimensions=None,
                     fill_value=None, withdata=True):
        """
        Wrapper on PseudoNetCDF.copyVariable that updates VAR-LIST,
        NVARS, VAR, and TFLAG

        See also
        --------
        see PseudoNetCDFFile.copyVariable
        """
        if key is None:
            for propk in ['_name', 'name', 'standard_name', 'long_name']:
                if hasattr(var, propk):
                    key = getattr(var, propk)
                    break
            else:
                raise AttributeError(
                    'varkey must be supplied because var has no name, ' +
                    'standard_name or long_name')

        outvar = PseudoNetCDFFile.copyVariable(
            self, var, key=key, dtype=dtype, dimensions=dimensions,
            fill_value=fill_value, withdata=withdata)

        # The long_name should be the same as the copied key
        outvar.long_name = key.ljust(16)
        # The var_desc and units should default to the copied variable
        if not hasattr(outvar, 'var_desc'):
            outvar.var_desc = key.ljust(80)
        if not hasattr(outvar, 'units'):
            outvar.units = 'unknown'.ljust(16)

        self._add2Varlist([key])
        return outvar

    def mask(self, *args, **kwds):
        """
        Wrapper on PseudoNetCDFFile.subsetVariables that updates VAR-LIST,
        NVARS, VAR, and TFLAG

        See also
        --------
        see PseudoNetCDFFile.mask
        """
        outf = PseudoNetCDFFile.mask(self, *args, **kwds)
        PseudoNetCDFFile.copyVariable(
            outf, self.variables['TFLAG'], key='TFLAG'
        )
        outf.updatemeta()
        return outf

    def subsetVariables(
        self, varkeys, inplace=False, exclude=False, keepcoords=True
    ):
        """
        Wrapper on PseudoNetCDFFile.subsetVariables that updates VAR-LIST,
        NVARS, VAR, and TFLAG

        See also
        --------
        see PseudoNetCDFFile.subsetVariables
        """
        varlist = self.getVarlist(update=False)
        newvarlist = [
            varkey for varkey in varlist
            if (
                (varkey in varkeys) != exclude and
                varkey in self.variables and
                not varkey.endswith('TFLAG')
            )
        ]
        outf = self.copy(props=True, dimensions=False, variables=False)
        for dk, dv in self.dimensions.items():
            if dk == 'VAR':
                outf.copyDimension(dv, key=dk, dimlen=len(newvarlist))
            else:
                outf.copyDimension(dv, key=dk)

        for vk in newvarlist:
            PseudoNetCDFFile.copyVariable(
                outf,
                self.variables[vk],
                key=vk,
                withdata=False
            )

        for vk in newvarlist:
            outf.variables[vk][:] = self.variables[vk][:]

        setattr(outf, 'VAR-LIST', '')
        outf._add2Varlist(newvarlist)
        outf.updatemeta()
        return outf

    def sliceDimensions(self, *args, **kwds):
        """
        Wrapper PseudoNetCDFFile.sliceDimensions that corrects ROW, COL,
        LAY and TIME meta-data according to the ioapi format

        Parameters
        ----------
        see PseudoNetCDFFile.sliceDimensions
        """
        # First slice as normal
        outf = PseudoNetCDFFile.sliceDimensions(self, *args, **kwds)
        # Copy slice keywords excluding newdims
        dimslices = kwds.copy()
        dimslices.pop('newdims', None)

        # Identify array indices and the need for fancy indexing
        isarray = {
            dk: not np.isscalar(dv) and not isinstance(dv, slice)
            for dk, dv in dimslices.items()
        }
        # anyisarray = np.sum(list(isarray.values())) > 1

        # Check if COL or ROW was used
        hascol = 'COL' in dimslices
        hasrow = 'ROW' in dimslices
        deleterowcol = False
        if hascol and hasrow:
            if isarray['ROW'] and isarray['COL']:
                newdims = kwds.get('newdims', ('POINTS',))
                if 'ROW' not in newdims and 'COL' not in newdims:
                    deleterowcol = True

        # If lay was subset, subset VGLVLS too
        if 'LAY' in kwds:
            nlvls = outf.VGLVLS.size
            lidx = np.array(
                np.arange(outf.VGLVLS.size - 1)[kwds['LAY']], ndmin=1
            )
            tmpvglvls = outf.VGLVLS[lidx]
            if lidx[-1] < (nlvls - 1):
                try:
                    endl = outf.VGLVLS[lidx[-1] + 1]
                    tmpvglvls = np.append(tmpvglvls, endl)
                    outf.VGLVLS = tmpvglvls
                except Exception:
                    warn('VGLVLS could not be diagnosed; update manually')
        # If subsetting replaces ('ROW', 'COL') ... for example with ('PERIM',)
        # remove the dimensions
        if deleterowcol:
            del outf.dimensions['COL']
            del outf.dimensions['ROW']
        else:
            # Update origins
            if 'COL' in kwds and 'COL' in outf.dimensions:
                ncol = len(self.dimensions['COL'])
                outf.XORIG += np.arange(ncol)[kwds['COL']].take(0) * outf.XCELL
            if 'ROW' in kwds and 'ROW' in outf.dimensions:
                nrow = len(self.dimensions['ROW'])
                outf.YORIG += np.arange(nrow)[kwds['ROW']].take(0) * outf.YCELL

        # Update TFLAG, SDATE, STIME and TSTEP
        if 'TSTEP' in kwds:
            import datetime
            times = np.atleast_1d(self.getTimes()[kwds['TSTEP']])
            outf.SDATE = int(times[0].strftime('%Y%j'))
            outf.STIME = int(times[0].strftime('%H%M%S'))
            if times.size > 1:
                dt = np.diff(times)
                if not (dt[0] == dt).all():
                    warn('New time is unstructured')
                outf.TSTEP = int(
                    (datetime.datetime(1900, 1, 1, 0) +
                     dt[0]).strftime('%H%M%S'))

        outf.updatemeta()
        return outf

    def interpSigma(self, vglvls, vgtop=None, interptype='linear',
                    extrapolate=False, fill_value='extrapolate',
                    verbose=0):
        """
        Parameters
        ----------
        vglvls : iterable
            the new vglvls (edges)
        vgtop : scalar
            Converting to new vgtop
        interptype : string
             'linear' uses a linear interpolation
             'conserve' uses a mass conserving interpolation
        extrapolate : boolean
            allow extrapolation beyond bounds with linear, default False
        fill_value : boolean
            set fill value (e.g, nan) to prevent extrapolation or edge
            continuation

        Returns
        -------
        outf : ioapi_base
            PseudoNetCDFFile with all variables interpolated

        Notes
        -----
        When extrapolate is false, the edge values are used for points beyond
        the inputs.
        """
        vglvls = np.asarray(vglvls)
        # grab input sigma coordinates
        myvglvls = self.VGLVLS

        # If needed, recalculate this files SIGMA
        if vgtop is not None and not vgtop == self.VGTOP:
            dp0 = 101325. - self.VGTOP
            dp1 = 101325. - vgtop
            myvglvls = (myvglvls * dp0 + self.VGTOP - vgtop) / dp1

        # Use midpoint for sigmas in inputs and outputs
        zs = (myvglvls[:-1] + myvglvls[1:]) / 2.
        nzs = (vglvls[:-1] + vglvls[1:]) / 2.

        if interptype == 'linear':
            from ..coordutil import getinterpweights
            weights = getinterpweights(
                zs, nzs, kind=interptype, fill_value=fill_value,
                extrapolate=extrapolate)
            # Create a function for interpolation

            def interpsigma(data):
                newdata = (weights * data[:, None]).sum(0)
                return newdata

        elif interptype == 'conserve':
            from ..coordutil import sigma2coeff
            # Calculate a weighting matrix using mass conserving
            # methods
            coeff = sigma2coeff(myvglvls, vglvls)  # (Nold, Nnew)
            # Calculate input mass fractions
            dp_in = -np.diff(myvglvls.astype('d'))[:, None]

            # Create a weighted mass fraction function
            fdp = dp_in * coeff

            # Create a precalculated normalizer
            ndp = fdp.sum(0)

            # Create a function for interpolation
            def interpsigma(data):
                nvals = (data[:, None] * fdp).sum(0) / ndp
                return nvals
        else:
            raise ValueError(
                'interptype only implemented for "linear" and "conserve"')

        # Apply function on LAY
        outf = self.applyAlongDimensions(LAY=interpsigma, verbose=verbose)

        # Ensure vglvls is a simple array
        outf.VGLVLS = vglvls.view(np.ndarray).astype('f')
        outf.NLAYS = len(outf.VGLVLS) - 1
        outf.updatemeta()
        return outf

    def _add2Varlist(self, varkeys):
        varliststr = getattr(self, 'VAR-LIST', '')
        keys = [k for k in varliststr.split() if k in self.variables]
        newkeys = set(varkeys).difference(keys + ['ETFLAG', 'TFLAG'])
        for varkey in varkeys:
            if varkey in newkeys:
                varliststr += varkey.ljust(16)
                keys.append(varkey)
        setattr(self, 'NVARS', len(keys))
        setattr(self, 'VAR-LIST', varliststr)
        self._updatetime()
        return keys

    def getVarlist(self, update=True, prune=True, retval='list'):
        """
        Returns
        -------
        update : boolean
            update files attributes to be consistent
        prune : bool
            If True, remove variables that are missing or have
            names longer than 16 characters
        retval : str
            Return formatted string 'str' or list otherwise

        Notes
        -----
        If VAR-LIST does not exist, it is added assuming all variables
        with dimensions ('TSTEP', 'LAY', ...) are variables
        """
        if getattr(self, 'VAR-LIST', ''):
            varliststr_old = ''
            varlist = [k for k, v in self.variables.items()
                       if v.dimensions[:2] == ('TSTEP', 'LAY')]
        else:
            varliststr_old = getattr(self, 'VAR-LIST')
            if len(varliststr_old) % 16 == 0:
                ivars = int(len(varliststr_old) // 16)
                varlist = [
                    varliststr_old[ivar * 16:ivar * 16 + 16].strip()
                    for ivar in range(ivars)
                ]
            else:
                varlist = varliststr_old.split()

        # To be in the VAR-LIST, the variable must be shorter than 16
        # and be present and have the typical IOAPI dimensions
        vardimchecks = {}
        for vk in varlist:
            if vk in self.variables:
                dims = tuple(self.variables[vk].dimensions)
            else:
                dims = ()
            check = (
                dims == ('TSTEP', 'LAY', 'ROW', 'COL')
                or dims == ('TSTEP', 'LAY', 'PERIM')
            )
            if check and len(vk) > 16:
                warn(
                    f'{vk} name is too long ({len(vk)});'
                    + ' not included in VAR-LIST'
                )
                check = False

            vardimchecks[vk] = check

        varlist = [
            vk for vk in varlist
            if prune and (
                vardimchecks[vk]
                and len(vk) <= 16
            )
        ]

        varliststr_new = ''.join([vk.ljust(16)[:16] for vk in varlist])
        if update and varliststr_new != varliststr_old:
            setattr(self, 'VAR-LIST', varliststr_new)
        if update and len(varlist) != self.NVARS:
            self.NVARS = len(varlist)

        newdimlen = max(self.NVARS, 1)
        if 'VAR' in self.dimensions:
            if newdimlen != len(self.dimensions['VAR']):
                try:
                    self.createDimension('VAR', newdimlen)
                except Exception:
                    pass
                # add updatetflag
        else:
            self.createDimension('VAR', newdimlen)

        if retval == 'str':
            return varliststr_new
        else:
            return varlist

    def updatetflag(self, overwrite=None, startdate=None, tstep=None):
        if overwrite is None:
            overwrite = (
                'TFLAG' not in self.variables or
                self.variables['TFLAG'].shape[1] != self.NVARS
            )

        if overwrite:
            if 'TFLAG' in self.variables:
                del self.variables['TFLAG']
            if startdate is not None:
                self.SDATE = int(startdate.strftime('%Y%j'))
                self.STIME = int(startdate.strftime('%H%M%S'))
            if tstep is not None:
                self.TSTEP = tstep

            times = self.getTimes()
            tvar = self.createVariable(
                'TFLAG', 'i', ('TSTEP', 'VAR', 'DATE-TIME'))
            tvar.units = '<YYYYDDD,HHMMSS>'.ljust(16)
            tvar.long_name = 'TFLAG'.ljust(16)
            tvar.var_desc = ("Timestep-valid flags:  (1) YYYYDDD or (2) " +
                             "HHMMSS                                ")

            yyyyjjj = np.array([int(t.strftime('%Y%j')) for t in times])
            hhmmss = np.array([int(t.strftime('%H%M%S')) for t in times])

            tvar[:, :, 0] = yyyyjjj[:, None].repeat(tvar.shape[1], 1)
            tvar[:, :, 1] = hhmmss[:, None].repeat(tvar.shape[1], 1)
            self.SDATE = tvar[0, 0, 0]
            self.STIME = tvar[0, 0, 1]
        else:
            if len(self.dimensions['VAR']) == 0:
                return

            try:
                times = self.getTimes()
            except Exception as e:
                warn('Times were incalculable: using epoch start\n' + str(e))
                times = np.array([datetime.datetime(1970, 1, 1)])

            if not hasattr(self, 'SDATE'):
                self.SDATE = int(times[0].strftime('%Y%j'))
            if not hasattr(self, 'STIME'):
                self.STIME = int(times[0].strftime('%H%M%S'))
            if not hasattr(self, 'TSTEP'):
                if times.size > 1:
                    dt = np.diff(times)
                    if not (dt[0] == dt).all():
                        warn('New time is unstructured')

                    tstep = int(
                        (
                            datetime.datetime(1900, 1, 1, 0) + dt.mean()
                        ).strftime('%H%M%S')
                    )
                    self.TSTEP = tstep
                else:
                    self.TSTEP = 10000

    def updatemeta(self, attdict={}, sortmeta=False):
        """
        Parameters
        ----------
        attdict : dictionary
            key value pairs to update meta data
        sortmeta : boolean
            sort meta data after update
        Returns
        -------
        None

        Notes
        -----
        Meta data not provided or present will be inferred or made up.
        (See _ioapi_defaults)
        """
        if attdict is None:
            attdict = {}
        attdict.update(_ioapi_defaults)
        for pk, pv in attdict.items():
            if not hasattr(self, pk):
                setattr(self, pk, pv)

        if 'TSTEP' in self.dimensions:
            td = self.dimensions['TSTEP']
            if not td.isunlimited():
                td.setunlimited(True)

        if 'DATE-TIME' not in self.dimensions:
            self.createDimension('DATE-TIME', 2)

        self.getVarlist(update=True)

        if 'LAY' in self.dimensions:
            self.NLAYS = len(self.dimensions['LAY'])
        if 'COL' in self.dimensions:
            self.NCOLS = len(self.dimensions['COL'])
        if 'ROW' in self.dimensions:
            self.NROWS = len(self.dimensions['ROW'])

        self._updatetime()
        self.updatetflag()
        if sortmeta:
            ioapi_sort_meta(self)

    def applyAlongDimensions(self, *args, **kwds):
        """
        Wrapper PseudoNetCDFFile.applyAlongDimensions that corrects ROW, COL,
        LAY and TIME meta-data according to the ioapi format

        Parameters
        ----------
        see PseudoNetCDFFile.applyAlongDimensions
        """
        outf = PseudoNetCDFFile.applyAlongDimensions(self, *args, **kwds)
        if 'LAY' in kwds:
            nlays = len(self.dimensions['LAY'])
            layf = PseudoNetCDFFile()
            layf.createDimension('lay', nlays)
            layf.createDimension('nv', 2)
            laym = layf.createVariable('lay', 'f', ('lay',))
            layb = layf.createVariable('lay_bounds', 'f', ('lay', 'nv'))
            layb[:, 0] = self.VGLVLS[:-1]
            layb[:, 1] = self.VGLVLS[1:]
            laym[:] = layb.mean(1)
            newlayf = layf.applyAlongDimensions(lay=kwds['LAY'])
            nlayb = newlayf.variables['lay_bounds']
            outf.VGLVLS = np.append(nlayb[:, 0], nlayb[:, 1]).view(np.ndarray)
        outf.updatemeta()
        return outf

    def eval(self, *args, **kwds):
        """
        Wrapper PseudoNetCDFFile.eval that corrects VAR-LIST
        and TFLAG meta-data according to the ioapi format

        Parameters
        ----------
        see PseudoNetCDFFile.eval
        """
        # oldkeys = set(self.variables)
        out = PseudoNetCDFFile.eval(self, *args, **kwds)
        outkeys = set(out.variables)
        # newkeys = outkeys.difference(oldkeys)
        # byekeys = oldkeys.difference(outkeys)
        out._add2Varlist(outkeys)
        out.updatemeta()
        return out

    def getMap(self, maptype='basemap_auto', **kwds):
        """
        Wrapper PseudoNetCDFFile.getMap that uses NCOLS, XCELL
        NROWS, and YCELL to calculate map bounds if basemap_auto

        Parameters
        ----------
        see PseudoNetCDFFile.getMap
        """
        if maptype.endswith('_auto'):
            if self.GDTYP == 1:
                lllon, lllat = self.XORIG, self.YORIG
                urlon = self.XORIG + self.NCOLS * self.XCELL
                urlat = self.YORIG + self.NROWS * self.YCELL
            else:
                lllon, lllat = self.xy2ll(0, 0)
                urlon, urlat = self.xy2ll(
                    self.NCOLS * self.XCELL, self.NROWS * self.YCELL)
            kwds.setdefault('llcrnrlon', lllon)
            kwds.setdefault('llcrnrlat', lllat)
            kwds.setdefault('urcrnrlon', urlon)
            kwds.setdefault('urcrnrlat', urlat)
            maptype = maptype[:-5]

        return PseudoNetCDFFile.getMap(self, maptype=maptype, **kwds)

    def plot(self, varkey, plottype=None, ax_kw=None,
             plot_kw=None, cbar_kw=None, map_kw=None, dimreduction='mean'):
        """
        Parameters
        ----------
        varkey : string
            the variable to plot
        plottype : string
            longitude-latitude, latitude-pressure, longitude-pressure,
            vertical-profile, time-longitude, time-latitude,
            time-pressure, default, last two dimensions in reverse order
        ax_kw : dictionary
            keywords for the axes to be created
        plot_kw : dictionary
            keywords for the plot (plot, scatter, or pcolormesh) to be
            created
        cbar_kw : dictionary or bool or None
            keywords for the colorbar; if True or None, use defaults.
            If False, do not create a colorbar
        map_kw : dictionary or bool or None
            keywords for the getMap routine, which is only used with
            map capable dimensions (ie, plottype='longitude-latitude')
            If True or None, use default configuration dict(countries=True,
            coastlines=True, states=False, counties=False). If False,
            do not draw a map.
        dimreduction : string or function
            dimensions not being used in the plot are removed
            using applyAlongDimensions(dimkey=dimreduction) where
            each dimenions.
        """

        import matplotlib.pyplot as plt
        from ..coordutil import getbounds

        if ax_kw is None:
            ax_kw = {}

        if plot_kw is None:
            plot_kw = {}

        if cbar_kw is None or cbar_kw is True:
            cbar_kw = {}

        if map_kw is None or map_kw is True:
            map_kw = {}

        apply2dim = {}
        var = self.variables[varkey]
        if plottype is None:
            vardims = var.dimensions
            if len(vardims) > 1:
                plottype = '-'.join(vardims[-2:][::-1])
            else:
                plottype = vardims[0] + '-profile'
        varunit = varkey
        if hasattr(var, 'units'):
            varunit += ' ' + var.units.strip()
        dimlens = dict([(dk, len(self.dimensions[dk]))
                        for dk in var.dimensions])
        dimpos = dict([(dk, di) for di, dk in enumerate(var.dimensions)])
        raw_xkey, raw_ykey = plottype.split('-')
        d2d = {'time': 'TSTEP', 'latitude': 'ROW',
               'longitude': 'COL', 'pressure': 'LAY'}
        xkey = d2d.get(raw_xkey, raw_xkey)
        ykey = d2d.get(raw_ykey, raw_ykey)
        if not ykey == 'profile':
            for dimkey in list(dimlens):
                if dimkey not in (xkey, ykey) and dimlens.get(dimkey, 1) > 1:
                    apply2dim[dimkey] = dimreduction

        if len(apply2dim) > 0:
            myf = self.applyAlongDimensions(**apply2dim)
            var = myf.variables[varkey]
            dimlens = dict([(dk, len(self.dimensions[dk]))
                            for dk in var.dimensions])
        else:
            myf = self
        if ykey in ('profile',):
            vaxi = var.dimensions.index(xkey)
            vsize = var.shape[vaxi]
            vals = np.rollaxis(var[:], vaxi).reshape(vsize, -1)
        else:
            vals = var[:].squeeze()

        if xkey == 'TSTEP':
            xm = myf.getTimes()
            dx = np.diff(xm)[-1]
            x = np.append(xm, xm[-1] + dx)
            x = plt.matplotlib.dates.date2num(x)
        else:
            x = getbounds(myf, xkey)

        ax = plt.gca(**ax_kw)
        if ykey in ('profile',):
            y = getbounds(myf, xkey)
            x1 = vals[:].min(1)
            xm = vals[:].mean(1)
            x2 = vals[:].max(1)
            ax.fill_betweenx(y=y, x1=x1, x2=x2, label=varkey + '(min, max)')
            ax.plot(xm, y, label=varkey, **plot_kw)
            ax.set_ylabel(xkey)
            ax.set_xlabel(varunit)
            return ax

        if ykey == 'TSTEP':
            ym = myf.getTimes()
            dy = np.diff(ym)[-1]
            y = np.append(ym, ym[-1] + dy)
            y = plt.matplotlib.dates.date2num(y)
        else:
            y = getbounds(myf, ykey)

        if dimpos[xkey] < dimpos[ykey]:
            vals = vals.T
        if xkey == 'TSTEP':
            ax.xaxis.set_major_formatter(
                plt.matplotlib.dates.AutoDateFormatter(
                    plt.matplotlib.dates.AutoDateLocator()))
        if ykey == 'TSTEP':
            ax.yaxis.set_major_formatter(
                plt.matplotlib.dates.AutoDateFormatter(
                    plt.matplotlib.dates.AutoDateLocator()))
        mappabledims = (
            plottype == 'longitude-latitude' or
            plottype == 'COL-ROW'
        )
        if mappabledims:
            if map_kw is not False:
                try:
                    coastlines = map_kw.pop('coastlines', True)
                    countries = map_kw.pop('countries', True)
                    states = map_kw.pop('states', False)
                    counties = map_kw.pop('counties', False)
                    bmap = myf.getMap(**map_kw)
                    if coastlines:
                        bmap.drawcoastlines(ax=ax)
                    if countries:
                        bmap.drawcountries(ax=ax)
                    if states:
                        bmap.drawstates(ax=ax)
                    if counties:
                        bmap.drawcounties(ax=ax)
                    x = np.arange(self.NCOLS + 1) * self.XCELL
                    y = np.arange(self.NROWS + 1) * self.YCELL
                    if self.GDTYP == 1:
                        x += self.XORIG
                        y += self.YORIG
                except Exception:
                    pass
        else:
            ax.set_xlabel(xkey)
            ax.set_ylabel(ykey)

        p = ax.pcolormesh(x, y, vals, **plot_kw)
        if cbar_kw is not False:
            cbar_kw.setdefault('label', varunit)
            ax.figure.colorbar(p, **cbar_kw)
        return ax

    slice = sliceDimensions
    apply = applyAlongDimensions


class ioapi(ioapi_base, netcdf):
    def __init__(self, *args, **kwds):
        netcdf.__init__(self, *args, **kwds)
        self.setCoords(['TFLAG'])

    def _newlike(self):
        if self.get_dest() is not None:
            outf = ioapi(**self.get_dest())
        elif isinstance(self, PseudoNetCDFFile):
            outt = ioapi_base
            outf = outt.__new__(outt)
        else:
            outf = PseudoNetCDFFile()
        outf.set_varopt(**self.get_varopt())
        outf._updatetime(write=True, create=True)
        return outf

    @classmethod
    def from_ncf(cls, infile):
        outf = ioapi_base()
        for pk in infile.ncattrs():
            pv = getattr(infile, pk)
            setattr(outf, pk, pv)

        for dk, dv in infile.dimensions.items():
            outf.copyDimension(dv, key=dk)

        for vk, vv in infile.variables.items():
            outf.copyVariable(vv, key=vk)

        return outf

    @property
    def _mode(self):
        return self.__dict__['_mode']

    def createVariable(self, *args, **kwds):
        return netcdf.createVariable(self, *args, **kwds)

    def createDimension(self, *args, **kwds):
        return netcdf.createDimension(self, *args, **kwds)

    @classmethod
    def isMine(cls, *args, **kwds):
        try:
            f = netcdf(*args, **kwds)
            for dk in ['TSTEP', 'VAR', 'DATE-TIME']:
                assert(dk in f.dimensions)
            attrlist = f.ncattrs()
            for pk in ['XORIG', 'XCELL', 'YCELL', 'YORIG', 'SDATE', 'STIME']:
                assert(pk in attrlist)
            return True
        except Exception:
            return False


def _fromdefaults(nfile, opts):
    for dk in opts:
        if dk in nfile.dimensions:
            return dk
    else:
        return opts[0]


def _tryset(obj, pk, pv, prefix='obj'):
    try:
        setattr(obj, pk, pv)
    except Exception as e:
        warn(
            '{}.{} not set {}; value {} was lost'.format(prefix, pk, e, pv)
        )


def ncf2ioapi(
    nfile, outpath, verbose=0, mode='w',
    tstepkeys=('TSTEP', 'Time', 'time', 't'),
    laykeys=('LAY', 'layer', 'level', 'layer47', 'layer72', 'z'),
    rowkeys=('ROW', 'latitude', 'south_north', 'lat', 'y'),
    colkeys=('COL', 'longitude', 'west_east', 'lon', 'x'),
    perimkeys=('PERIM', 'POINTS'),
    format='NETCDF3_CLASSIC',
    **props
):
    """
    Parameters
    ----------
    nfile : netcdf-like file
        Input file to translate to IOAPI meta data NETCDF file
    outpath : str
        Path for output file
    verbose : int
        Higher verbosity is more info
    mode : str
        Mode for opening file (w, ws, r+, a)
    tstepkeys : tuple of strings
        Strings to translate to TSTEP
    laykeys : tuple of strings
        Strings to translate to LAY
    rowkeys : tuple of strings
        Strings to translate to ROW
    colkeys : tuple of strings
        Strings to translate to COL
    perimkeys : tuple of strings
        Strings to translate to PERIM; only relevant when GDTYP=2
    format : string
        Format for output file either NETCDF3_CLASSIC, NETCDF3_64BIT_OFFSET
        NETCDF4_CLASSIC
    props : dict
        Mapping of properties to set

    Returns
    -------
    out : netCDF4.Dataset
    """
    time = _fromdefaults(nfile, tstepkeys)
    lay = _fromdefaults(nfile, laykeys)
    row = _fromdefaults(nfile, rowkeys)
    col = _fromdefaults(nfile, colkeys)
    perim = _fromdefaults(nfile, perimkeys)
    dim2dim = {}
    if time != 'TSTEP':
        dim2dim[time] = 'TSTEP'
    if lay != 'LAY':
        dim2dim[lay] = 'LAY'
    if row != 'ROW':
        dim2dim[row] = 'ROW'
    if col != 'COL':
        dim2dim[col] = 'COL'
    if col != 'COL':
        dim2dim[col] = 'COL'

    if verbose > 0:
        print('Dimension translation', dim2dim)

    fileprops = _ioapi_defaults.copy()
    for k in nfile.ncattrs():
        fileprops[k] = getattr(nfile, k)

    fileprops.update(**props)

    if fileprops['FTYPE'] == 2:
        indims = (time, lay, perim)
    else:
        indims = (time, lay, row, col)

    outdims = tuple(dim2dim.get(k, k) for k in indims)
    varkeys = [k for k, v in nfile.variables.items() if v.dimensions == indims]
    varlist = ''.join([k.ljust(16) for k in varkeys])

    fileprops['VAR-LIST'] = varlist
    nvar = fileprops['NVARS'] = len(varkeys)
    nthk = fileprops['NTHIK']
    nx = fileprops['NCOLS'] = len(nfile.dimensions[col])
    ny = fileprops['NROWS'] = len(nfile.dimensions[row])
    nz = fileprops['NLAYS'] = len(nfile.dimensions[lay])

    ofile = netcdf(outpath, format=format, mode=mode)

    for pk, pv in fileprops.items():
        _tryset(ofile, pk, pv, prefix='file')

    ofile.createDimension('TSTEP', None)
    ofile.createDimension('DATE-TIME', 2)
    ofile.createDimension('LAY', nz)
    ofile.createDimension('VAR', nvar)

    if fileprops['FTYPE'] == 2:
        ofile.createDimension('PERIM', nthk * (4 * nthk + 2 * (nx + ny)))
    elif fileprops['FTYPE'] == 1:
        ofile.createDimension('ROW', ny)
        ofile.createDimension('COL', nx)
    else:
        raise ValueError('FTYPE is unknown; must be 1 or 2')

    times = nfile.getTimes()
    tv = ofile.createVariable('TFLAG', 'i', ('TSTEP', 'VAR', 'DATE-TIME'))
    tv.units = '<YYYYJJJ,HHMMSS>'
    tv.long_name = 'TFLAG'.ljust(16)
    tv.var_desc = 'TFLAG'.ljust(80)
    yyyyjjj = np.array([t.strftime('%Y%j') for t in times], dtype='i')
    hhmmss = np.array([t.strftime('%H%M%S') for t in times], dtype='i')
    for k in varkeys:
        v = nfile.variables[k]
        ov = ofile.createVariable(k, v.dtype.char, outdims)
        ov.long_name = k.ljust(16)
        ov.var_desc = k.ljust(16)
        ov.units = 'unknown'.ljust(16)
        for pk in v.ncattrs():
            pv = getattr(v, pk)
            _tryset(ov, pk, pv, prefix=k)

    tv[:yyyyjjj.size, :, 0] = yyyyjjj[:, None].repeat(nvar, 1)
    tv[:yyyyjjj.size, :, 1] = hhmmss[:, None].repeat(nvar, 1)
    dt = (times[-1] - times[0]).total_seconds() / (len(times) - 1)
    tmpd = datetime.datetime(1900, 1, 1) + datetime.timedelta(seconds=dt)
    ofile.TSTEP = int(tmpd.strftime('%H%M%S'))
    ofile.SDATE = int(times[0].strftime('%Y%j'))
    ofile.STIME = int(times[0].strftime('%H%M%S'))
    # if (
    #     any([vk.startswith('lat') for vk in nfile.variables]) and
    #     any([vk.startswith('lon') for vk in nfile.variables])
    # ):
    #     ofile.GDTYP = 1
    #     projstr = nfile.getproj(projformat='proj4', withgrid=True)
    if not hasattr(ofile, 'VGLVLS'):
        ofile.VGTYP = -1
        ofile.VGLVLS = np.arange(nz + 1, dtype='f')
    for k in varkeys:
        v = nfile.variables[k]
        ov = ofile.variables[k]
        ov[:] = v

    return ofile


registerwriter('ioapi', ncf2ioapi)



__doc__ = """
.. _CAMx
:mod:`CAMx` -- CAMx File Interfaces
===================================

.. module:: CAMx
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map and random access read
              based file interfaces for all CAMx related files.  See
              PseudoNetCDF.sci_var.PseudoNetCDFFile for interface details.
              Each camx file level module (i.e. uamiv, cloud_rain, etc) can
              be called as a script for ncdump style output.  Example:
              python -m PseudoNetCDF.camxfiles.uamiv
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
__all__ = ['Memmaps',
           'Readers',
           'Writers',
           'cloud_rain',
           'height_pressure',
           'humidity',
           'ipr',
           'irr',
           'landuse',
           'lateral_boundary',
           'one3d',
           'point_source',
           'temperature',
           'uamiv',
           'vertical_diffusivity',
           'wind',
           'FortranFileUtil',
           'util',
           'timetuple']

from . import cloud_rain
from . import height_pressure
from . import humidity
from . import ipr
from . import irr
from . import landuse
from . import one3d
from . import point_source
from . import temperature
from . import uamiv
from . import vertical_diffusivity
from . import wind
from . import FortranFileUtil
from . import util
from . import timetuple
from . import lateral_boundary
from . import Readers
from . import Memmaps
from . import Writers



#!/usr/bin/env python
from __future__ import print_function

from scipy.constants import F2K
import os
import json
import datetime
try:
    from urllib.request import urlopen
except Exception:
    from urllib2 import urlopen


# IEM quirk to have Iowa AWOS sites in its own labeled network
from shapely.geometry import Polygon, Point
import numpy as np
import argparse
from netCDF4 import Dataset
parser = argparse.ArgumentParser(
    prog='METAR',
    description="Downloads met data from " +
                "http://mesonet.agron.iastate.edu for a CMAQ domain")
parser.add_argument('--network', default='US__ASOS',
                    help='Network follwing country_state_type ' +
                         'conventions e.g., US_FL_ASOS')
spacegroup = parser.add_mutually_exclusive_group(required=True)
spacegroup.add_argument('--gridcro2d', dest='GRIDCRO2D',
                        help='CMAQ MCIP GRIDCRO2D file or any ' +
                             'file that has LAT and LON variables')
spacegroup.add_argument('--wktpolygon')
parser.add_argument('--START_DATE', default=None, help='YYYY-MM-DD')
parser.add_argument('--END_DATE', default=None, help='YYYY-MM-DD')
parser.add_argument('outpath', help='Root outpath')
args = parser.parse_args()
print(args)

if args.GRIDCRO2D is not None:
    f = Dataset(args.GRIDCRO2D)
    lon = f.variables['LON'][0, 0]
    lat = f.variables['LAT'][0, 0]
    llcrnrlon = lon[:, 0].max()
    urcrnrlon = lon[:, -1].min()
    llcrnrlat = lat[0, :].max()
    urcrnrlat = lat[-1, :].min()
    bounds = Polygon([(llcrnrlon, llcrnrlat), (llcrnrlon, urcrnrlat),
                      (urcrnrlon, urcrnrlat), (urcrnrlon, llcrnrlat)])
else:
    from shapely.wkt import loads
    bounds = loads(args.wktpolygon)

if args.START_DATE is None:
    startts = datetime.datetime.strptime(str(f.SDATE), '%Y%j')
else:
    startts = datetime.datetime.strptime(args.START_DATE, '%Y-%m-%d')
if args.END_DATE is None:
    endts = datetime.datetime.strptime(str(f.SDATE), '%Y%j')
else:
    endts = datetime.datetime.strptime(args.END_DATE, '%Y-%m-%d')


# timestamps in UTC with latlon in tab delimited format to request data for

SERVICE = "http://mesonet.agron.iastate.edu/cgi-bin/request/asos.py?"
SERVICE += "data=all&tz=Etc/UTC&format=tdf&latlon=yes&"

SERVICE += startts.strftime('year1=%Y&month1=%m&day1=%d&')
SERVICE += endts.strftime('year2=%Y&month2=%m&day2=%d&')

sites = []
sitelocs = {}
for network in args.network.split(','):
    # Get metadata
    uri = "http://mesonet.agron.iastate.edu/geojson/network.php?network=%s" % (
        network,)
    data = urlopen(uri).read().decode('ascii')
    jdict = json.loads(data)

    nsites = [f['properties']['sid'] for f in jdict['features']
              if bounds.intersects(Point(f['geometry']['coordinates']))]
    nlocs = dict([(f['properties']['sid'], f['geometry']['coordinates'])
                  for f in jdict['features']
                  if bounds.intersects(Point(f['geometry']['coordinates']))])
    print(nsites)
    for site in nsites:
        sitelocs[site] = nlocs[site]
        uri = '%s&station=%s' % (SERVICE, site)
        print('Network: %s Downloading: %s' % (network, site))
        outfn = '%s_%s_%s.txt' % (site, startts.strftime("%Y%m%d%H%M"),
                                  endts.strftime("%Y%m%d%H%M"))
        if os.path.exists(outfn):
            print('Using cached')
            continue
        data = urlopen(uri)
        out = open(outfn, 'w')
        out.write(data.read().decode('ascii'))
        out.close()
    sites += nsites


outf = Dataset(args.outpath, mode='w')
outf.createDimension('TSTEP', None)
outf.createDimension('VAR', 4)
outf.createDimension('DATE-TIME', 2)
outf.createDimension('tnv', 2)
outf.createDimension('LAY', 1)
outf.createDimension('points', len(sites))
outf.SDATE = int(startts.strftime('%Y%j'))
outf.STIME = 0
tmpnhours = int((endts - startts).total_seconds() // 3600) + 24
time_bounds = [(startts + datetime.timedelta(hours=i),
                startts + datetime.timedelta(hours=i + 1))
               for i in range(tmpnhours)]

tflag = outf.createVariable('TFLAG', 'i', ('TSTEP', 'VAR', 'DATE-TIME'))
nstep = len(time_bounds)
tflag[0:nstep, :, 0] = 0
tflag[0:nstep, :, 0] = np.array(
    [int(t[0].strftime('%Y%j')) for t in time_bounds])[:, None].repeat(4, 1)
tflag[0:nstep, :, 1] = np.array(
    [int(t[0].strftime('%H%M%S')) for t in time_bounds])[:, None].repeat(4, 1)

rtime = datetime.datetime(1970, 1, 1, 0)
time = outf.createVariable('time', 'd', ('TSTEP',), fill_value=-999.)
time[:] = np.array([(ts - rtime + datetime.timedelta(hours=.5)
                     ).total_seconds() for ts, te in time_bounds])
time.units = 'seconds since 1970-01-01 00:00:00 UTC'

time_bounds_var = outf.createVariable(
    'time_bounds', 'd', ('TSTEP', 'tnv'), fill_value=-999.)
time_bounds_var[:, 0] = np.array(
    [(ts - rtime).total_seconds() for ts, te in time_bounds])
time_bounds_var[:, 1] = np.array(
    [(te - rtime).total_seconds() for ts, te in time_bounds])
time_bounds_var.units = 'seconds since 1970-01-01 00:00:00 UTC'

temp2 = outf.createVariable(
    'TEMP2', 'f', ('TSTEP', 'LAY', 'points'), fill_value=-999)
temp2.units = 'K'.ljust(16)

wdir10 = outf.createVariable(
    'WDIR10', 'f', ('TSTEP', 'LAY', 'points'), fill_value=-999)
wdir10.units = 'degrees from N'.ljust(16)

wspd10 = outf.createVariable(
    'WSPD10', 'f', ('TSTEP', 'LAY', 'points'), fill_value=-999)
wspd10.units = 'M/S'.ljust(16)

prsfc = outf.createVariable(
    'PRSFC', 'f', ('TSTEP', 'LAY', 'points'), fill_value=-999)
prsfc.units = 'Pa'.ljust(16)


rn = outf.createVariable(
    'RN', 'f', ('TSTEP', 'LAY', 'points'), fill_value=-999)
rn.units = 'CM'.ljust(16)

lat = outf.createVariable('LAT', 'f', ('points'))
lat.units = 'degrees_north'.ljust(16)
lon = outf.createVariable('LON', 'f', ('points'))
lon.units = 'degrees_east'.ljust(16)

for si, site in enumerate(sites):
    outfn = '%s_%s_%s.txt' % (site, startts.strftime("%Y%m%d%H%M"),
                              endts.strftime("%Y%m%d%H%M"))
    data = np.recfromtxt(outfn, missing_values='M', delimiter='\t',
                         names=True, comments='#', skip_header=5, usemask=True)
    lat[si] = sitelocs[site][1]
    lon[si] = sitelocs[site][0]
    if data.size == 0:
        continue
    times = np.array([datetime.datetime.strptime(
        d.decode('ascii'), '%Y-%m-%d %H:%M') for d in data['valid']])
    for ti, (sh, eh) in enumerate(time_bounds):
        tdata = data[(times > sh) & (times <= eh)]
        if 'tmpf' in data.dtype.names:
            val = F2K(tdata['tmpf']).mean()
            if not np.ma.is_masked(val):
                temp2[ti, 0, si] = val
        if 'drct' in data.dtype.names:
            val = tdata['drct'].mean()
            if not np.ma.is_masked(val):
                wdir10[ti, 0, si] = val
        if 'sknt' in data.dtype.names:
            val = tdata['sknt'].mean() * 0.514444444  # convert knot to m/s
            if not np.ma.is_masked(val):
                wspd10[ti, 0, si] = val
        if 'alti' in data.dtype.names:
            val = tdata['alti'].mean() * 760 / 29.9213 * 101325. / \
                760.  # convert inches to Pa
            if not np.ma.is_masked(val):
                prsfc[ti, 0, si] = val
        if 'p01i' in data.dtype.names:
            val = tdata['p01i'].mean() * 2.54  # convert inches to cm
            if not np.ma.is_masked(val):
                rn[ti, 0, si] = val
    outf.sync()
outf.close()



__all__ = ['PseudoNetCDFFile', 'netcdf', 'PseudoNetCDFVariables',
           'PseudoNetCDFVariable', 'PseudoNetCDFMaskedVariable',
           'PseudoIOAPIVariable', 'PseudoNetCDFDimension',
           'pncrename', 'manglenames', 'removesingleton', 'getvarpnc',
           'interpvars', 'extract_from_file', 'extract_lonlat', 'mask_vals',
           'slice_dim', 'reduce_dim', 'pncfunc', 'pncbo', 'pncbfunc',
           'pncexpr', 'seqpncbo', 'mesh_dim', 'add_attr', 'convolve_dim',
           'merge', 'stack_files', 'splitdim']

from ._files import PseudoNetCDFFile, netcdf, PseudoNetCDFVariables
from ._variables import PseudoNetCDFVariable, PseudoNetCDFMaskedVariable
from ._variables import PseudoIOAPIVariable
from ._functions import pncrename, manglenames, removesingleton, getvarpnc
from ._functions import interpvars, extract_from_file, extract_lonlat
from ._functions import mask_vals, slice_dim, reduce_dim, pncfunc
from ._functions import pncbo, pncbfunc, pncexpr, seqpncbo, mesh_dim
from ._functions import add_attr, convolve_dim, merge, stack_files, splitdim
from ._dimensions import PseudoNetCDFDimension



__all__ = ['write_point']
__doc__ = """
.. _Write
:mod:`Write` -- CAMx point_source writer
============================================

.. module:: Write
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` writer for CAMx
              point_source files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

# Distribution packages
import operator
from functools import reduce

# Site-Packages
import numpy as np

# This Package modules
from PseudoNetCDF.camxfiles.timetuple import timeadd, timerange
from PseudoNetCDF.camxfiles.FortranFileUtil import writeline, Asc2Int
from PseudoNetCDF._getwriter import registerwriter

_emiss_hdr_fmt = np.dtype(dict(
    names=['SPAD', 'name', 'note', 'itzon', 'nspec', 'ibdate', 'btime',
           'iedate', 'etime', 'EPAD'],
    formats=['>i', '(10,4)>S1', '(60,4)>S1', '>i', '>i', '>i', '>f', '>i',
             '>f', '>i']))

_grid_hdr_fmt = np.dtype(dict(
    names=['SPAD', 'plon', 'plat', 'iutm', 'xorg', 'yorg', 'delx', 'dely',
           'nx', 'ny', 'nz', 'iproj', 'istag', 'tlat1', 'tlat2', 'rdum5',
           'EPAD'],
    formats=['>i', '>f', '>f', '>i', '>f', '>f', '>f', '>f', '>i', '>i',
             '>i', '>i', '>i', '>f', '>f', '>f', '>i']))

_cell_hdr_fmt = np.dtype(dict(
    names=['SPAD', 'ione1', 'ione2', 'nx', 'ny', 'EPAD'],
    formats=['>i', '>i', '>i', '>i', '>i', '>i']))

_time_hdr_fmt = np.dtype(dict(
    names=['SPAD', 'ibdate', 'btime', 'iedate', 'etime', 'EPAD'],
    formats=['>i', '>i', '>f', '>i', '>f', '>i']))


_nstk_hdr_fmt = np.dtype(dict(
    names=['SPAD', 'ione', 'nstk', 'EPAD'],
    formats=['>i', '>i', '>i', '>i']))


def ncf2point_source(ncffile, outpath):
    outfile = open(outpath, 'wb')
    emiss_hdr = np.zeros((1,), dtype=_emiss_hdr_fmt)
    emiss_hdr['SPAD'] = emiss_hdr['EPAD'] = emiss_hdr.dtype.itemsize - 8
    emiss_hdr[0]['name'][:, :] = ' '
    emiss_hdr[0]['name'][:, 0] = np.array(
        [c for c in ncffile.NAME], dtype='>S1')
    emiss_hdr[0]['note'][:, :] = ' '
    emiss_hdr[0]['note'][:, 0] = np.array(
        [c for c in ncffile.NOTE], dtype='>S1')
    # gdtype = getattr(ncffile, 'GDTYPE', -999)
    emiss_hdr['itzon'][0] = ncffile.ITZON
    tmps1 = np.array([c for c in ncffile.SPC_NAMES], dtype='S1')
    tmps1 = tmps1.reshape(-1, 16)[:, :10].copy().view('>S10')[:, 0]
    SPC_NAMES = [s.decode() for s in tmps1]
    nspec = len(SPC_NAMES)
    emiss_hdr['nspec'] = nspec

    grid_hdr = np.zeros((1,), dtype=_grid_hdr_fmt)
    grid_hdr['SPAD'] = grid_hdr['EPAD'] = grid_hdr.dtype.itemsize - 8
    grid_hdr['plon'] = ncffile.PLON
    grid_hdr['plat'] = ncffile.PLAT
    grid_hdr['iutm'][0] = ncffile.IUTM
    grid_hdr['xorg'] = ncffile.XORIG
    grid_hdr['yorg'] = ncffile.YORIG
    grid_hdr['delx'] = ncffile.XCELL
    grid_hdr['dely'] = ncffile.YCELL
    grid_hdr['nx'] = ncffile.NCOLS
    grid_hdr['ny'] = ncffile.NROWS
    grid_hdr['nz'] = ncffile.NLAYS
    grid_hdr['iproj'] = ncffile.CPROJ
    grid_hdr['tlat1'] = ncffile.TLAT1
    grid_hdr['tlat2'] = ncffile.TLAT2
    grid_hdr['istag'] = ncffile.ISTAG
    grid_hdr['rdum5'] = 100.

    cell_hdr = np.zeros((1,), dtype=_cell_hdr_fmt)
    cell_hdr['SPAD'] = cell_hdr['EPAD'] = cell_hdr.dtype.itemsize - 8
    cell_hdr['ione1'] = 1
    cell_hdr['ione2'] = 1
    cell_hdr['nx'] = ncffile.NCOLS
    cell_hdr['ny'] = ncffile.NROWS
    time_hdr = np.zeros(
        shape=(len(ncffile.dimensions['TSTEP']),), dtype=_time_hdr_fmt)
    time_hdr['SPAD'] = time_hdr['EPAD'] = time_hdr.dtype.itemsize - 8
    date, time = ncffile.variables['TFLAG'][:, 0].T
    edate, etime = ncffile.variables['ETFLAG'][:, 0].T
    time = time.astype('>f') / 10000.
    date = date % (date // 100000 * 100000)
    etime = etime.astype('>f') / 10000.
    edate = edate % (edate // 100000 * 100000)
    time_hdr['ibdate'] = date
    time_hdr['btime'] = time
    time_hdr['iedate'] = edate
    time_hdr['etime'] = etime
    # time_hdr['iedate'] += time_hdr['etime'] // 24
    # time_hdr['etime'] -= (time_hdr['etime'] // 24) * 24

    emiss_hdr['ibdate'] = time_hdr[0]['ibdate']
    emiss_hdr['btime'] = time_hdr[0]['btime']
    emiss_hdr['iedate'] = ncffile.EDATE
    emiss_hdr['etime'] = ncffile.ETIME

    nspec = emiss_hdr['nspec'][0]
    _spc_hdr_fmt = np.dtype(dict(names=['SPAD', 'NAME', 'EPAD'], formats=[
                            '>i', '(%d,10,4)>S1' % nspec, '>i']))
    spc_hdr = np.zeros((1,), dtype=_spc_hdr_fmt)
    spc_hdr['NAME'][:] = ' '
    spc_hdr['SPAD'] = spc_hdr['EPAD'] = spc_hdr.dtype.itemsize - 8
    for si, spc in enumerate(SPC_NAMES):
        spc_hdr[0]['NAME'][si, :, 0] = np.array(
            [c for c in spc.ljust(10)], dtype='>S1')

    nstk_hdr = np.ones((1,), dtype=_nstk_hdr_fmt)
    nstk_hdr['SPAD'] = nstk_hdr['EPAD'] = nstk_hdr.dtype.itemsize - 8
    nstk_hdr['nstk'] = len(ncffile.dimensions['NSTK'])

    nstk = nstk_hdr['nstk'][0]
    # (xstk(n),ystk(n),hstk(n),dstk(n),tstk(n),vstk(n),n=1,nstk)
    stk_prop_fmt = np.dtype(dict(
        names=['xstk', 'ystk', 'hstk', 'dstk', 'tstk', 'vstk'],
        formats=['>f', '>f', '>f', '>f', '>f', '>f']))
    stk_prop = np.ones((1,), dtype=np.dtype([
        ('SPAD', '>i', (1,)), ('DATA', stk_prop_fmt, nstk),
        ('EPAD', '>i', (1,))
    ]))
    stk_prop['SPAD'] = stk_prop['EPAD'] = stk_prop.dtype.itemsize - 8
    stk_prop['DATA']['xstk'] = ncffile.variables['XSTK'][:]
    stk_prop['DATA']['ystk'] = ncffile.variables['YSTK'][:]
    stk_prop['DATA']['hstk'] = ncffile.variables['HSTK'][:]
    stk_prop['DATA']['dstk'] = ncffile.variables['DSTK'][:]
    stk_prop['DATA']['tstk'] = ncffile.variables['TSTK'][:]
    stk_prop['DATA']['vstk'] = ncffile.variables['VSTK'][:]

    # Start writing
    emiss_hdr.tofile(outfile)
    grid_hdr.tofile(outfile)
    cell_hdr.tofile(outfile)
    spc_hdr.tofile(outfile)
    nstk_hdr.tofile(outfile)
    spc_names = [spc[:, 0].copy().view('>S10')[0].decode().strip()
                 for spc in spc_hdr[0]['NAME']]
    stk_prop.tofile(outfile)
    ione = np.array([1]).astype('>i')
    for di, (d, t) in enumerate(ncffile.variables['TFLAG'][:, 0]):
        tempout = b''
        tempout += time_hdr[di].tobytes()
        tempout += np.array([8, 1, nstk, 8], dtype='>i').tobytes()
        # (idum,idum,kcell(n),flow(n),plmht(n),n=1,nstk)
        time_prop_fmt = np.dtype(dict(
            names=['ione1', 'ione2', 'kcell', 'flow', 'plmht'],
            formats=['>i', '>i', '>i', '>f', '>f']))
        props = np.ones((1,), dtype=np.dtype(
            [('SPAD', '>i', (1,)),
             ('DATA', time_prop_fmt, (nstk,)),
             ('EPAD', '>i', (1,))]))
        props['SPAD'] = props['EPAD'] = props.dtype.itemsize - 8
        props['DATA']['ione1'] = ncffile.variables['IONE'][di]
        props['DATA']['ione2'] = ncffile.variables['ITWO'][di]
        props['DATA']['kcell'] = ncffile.variables['KCELL'][di]
        props['DATA']['flow'] = ncffile.variables['FLOW'][di]
        props['DATA']['plmht'] = ncffile.variables['PLMHT'][di]
        tempout += props.tobytes()
        for spc_key, spc_name in zip(spc_names, spc_hdr[0]['NAME']):
            var = ncffile.variables[spc_key]
            data = var[di].astype('>f')
            buf = np.array([4 + 40 + data.size * 4]).astype('>i')
            tempout += buf.tobytes()
            tempout += ione.tobytes()
            tempout += spc_name.tobytes()
            tempout += data.tobytes()
            tempout += buf.tobytes()
        outfile.write(tempout)
    outfile.flush()
    return outfile


registerwriter('camxfiles.point_source', ncf2point_source)
registerwriter('point_source', ncf2point_source)


def write_point(start_date, start_time, time_step, hdr, vals):
    # Internalize header
    hdr = [h for h in hdr]
    species = hdr[-4]
    nstk = hdr[-3][1]
    timeprops = hdr.pop()

    # initialize hdr_fmts with species count
    hdr_fmts = ["10i60i3ifif", "ffiffffiiiiifff",
                "iiii", "10i" * len(species), "ii", "ffffff" * nstk]

    # initialize output variable
    pt_string = ''

    # Change name and note
    hdr[0] = list(hdr[0][0]) + list(hdr[0][1]) + list(hdr[0][2:])

    # Reducing stk props
    stkprops = hdr[-1]
    hdr[-1] = []
    for stk in stkprops:
        hdr[-1].extend(stk)
    stk_time_prop_fmt = "iiiff" * nstk

    stk_time_props = []
    for time in timeprops:
        stk_time_props.append([])
        for stk in time:
            stk_time_props[-1].extend(stk)

    # Change species names to array of characters
    hdr[-3] = reduce(operator.concat, [Asc2Int(s) for s in hdr[-3]])

    # for each item in the header, write it to output
    for i, (h, f) in enumerate(zip(hdr, hdr_fmts)):
        pt_string += writeline(h, f, False)

    # create value format
    valfmt = 'i10i' + ('f' * nstk)

    # Get end date
    (end_date, end_time) = timeadd(
        (start_date, start_time), (0, time_step * vals.shape[0]))

    # Write out values
    for ti, (d, t) in enumerate(timerange((start_date, start_time),
                                          (end_date, end_time),
                                          time_step)):
        ed, et = timeadd((d, t), (0, time_step))
        pt_string += writeline((d, t, ed, et), 'ifif', False)
        pt_string += writeline((1, nstk), 'ii', False)
        pt_string += writeline(stk_time_props[ti], stk_time_prop_fmt, False)
        for si, spc in enumerate(species):
            # Dummy variable,spc characters and values flattened
            temp = [1]
            temp.extend(Asc2Int(spc))
            temp.extend(vals[ti, si, ...])
            pt_string += writeline(temp, valfmt, False)

    return pt_string



"""
Written to read files from http://croc.gsfc.nasa.gov/shadoz/
"""
from PseudoNetCDF import PseudoNetCDFFile
import re
import numpy as np
spaces = re.compile(r'\s{2,1000}')


def shadoz(inpath):
    datafile = open(inpath, 'r')
    datalines = datafile.read().split('\n')
    nmeta = int(datalines[0])
    meta = dict([[w.strip() for w in l.split(': ')]
                 for l in datalines[1:nmeta - 2]])
    varline, unitline = datalines[nmeta - 2:nmeta]
    varnames = spaces.split(varline)
    units = spaces.split(unitline)
    data = np.fromstring(
        '\n'.join(datalines[nmeta:]), sep=' ').reshape(-1, len(varnames))
    outf = PseudoNetCDFFile()
    outf.createDimension('time', data.shape[0])
    missing = -9999
    for k, v in meta.items():
        setattr(outf, k, v)
        if k == 'Missing or bad values':
            missing = eval(v)

    for varname, unit, vals in zip(varnames, units, data.T):
        var = outf.createVariable(varname.replace(' ', '_'), 'f', ('time',))
        var.units = unit
        var.standard_name = varname
        var[:] = np.ma.masked_values(vals[:], missing)

    return outf


if __name__ == '__main__':
    f = shadoz('/Users/barronh/Downloads/costarica_20160108_V05.1_R.dat')



__doc__ = """
.. _one3d
:mod:`one3d` -- one3d File Interfaces
=====================================

.. module:: one3d
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map and random access read
   based file interfaces for CAMx one variable 3d files.
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
__all__ = ['Memmap', 'Read', 'Write']

from . import Memmap
from . import Read
from . import Write



__all__ = ['l100']
from datetime import datetime
from PseudoNetCDF import PseudoNetCDFFile
import numpy as np

_edges = [0, 4, 12, 20, 28, 35, 43, 50, 57, 64, 72, 79, 88, 95]
_starts = _edges[:-1]
_ends = _edges[1:]
_ses = [se for se in zip(_starts, _ends)]
_orignames = ('Level Press Alt Pottp Temp FtempV Hum Ozone Ozone ' +
              'Ozone Ptemp O3_DN O3_Res').split()
_names = ('Level Press Alt Pottp Temp FtempV Hum Ozone_mPa Ozone_ppmv ' +
          'Ozone_atmcm Ptemp O3_DN O3_Res').split()
_units = 'Num hPa km K C C % mPa ppmv atmcm C 10^11/cc DU'.split()


class l100(PseudoNetCDFFile):
    @classmethod
    def isMine(cls, path):
        """
        Must have correct names on line 28
        """
        try:
            lines = cls._getmeta(path)
            mynames = lines[-2].split()
            for chk, new in zip(_orignames[:8], mynames):
                if chk != new:
                    return False
            else:
                return True
        except Exception:
            return False

    @classmethod
    def _getmeta(cls, path):

        # get metadata lines
        infile = open(path)
        lines = []
        for li in range(100):
            line = infile.readline()
            lines.append(line)
            if line.startswith('Level'):
                lines.append(infile.readline())
                break
        else:
            lines = lines[:29]
        return lines

    def __init__(self, path):
        try:
            import pandas as pd
        except Exception:
            raise ImportError('l100 sonde files requires pandas; ' +
                              'install pandas (e.g., pip install pandas)')
        self._path = path
        metalines = l100._getmeta(path)
        datastartline = len(metalines)
        for line in metalines:
            if ':' in line:
                parts = line.split(':')
                key = parts[0].strip().replace(' ', '_')
                val = ':'.join(parts[1:]).strip()
                setattr(self, key, val)

        dataset = pd.read_fwf(path, colspecs=_ses, skiprows=datastartline,
                              names=_names, na_values=['99.90', '99.99', '999',
                                                       '99.999', '999.999',
                                                       '99.9990', '9999'])
        self.createDimension('site', 1)
        self.createDimension('level', dataset.shape[0])
        for key, unit in zip(_names, _units):
            vals = dataset[key]
            outvar = self.createVariable(key, vals.dtype, ('site', 'level'))
            outvar.long_name = key
            outvar.units = unit
            outvar[0, :] = vals

        ldate = datetime.strptime(
            self.Launch_Date + ' ' + self.Launch_Time, '%d %B %Y %H:%M:%S %Z')
        rdate = datetime(1970, 1, 1)
        seconds = (ldate - rdate).total_seconds()
        lonvar = self.createVariable('longitude', 'd', ('site',))
        lonvar.long_name = 'longitude'
        lonvar.units = 'degrees'
        lonvar[:] = eval(self.Longitude)
        latvar = self.createVariable('latitude', 'd', ('site',))
        latvar.long_name = 'latitude'
        latvar.units = 'degrees'
        latvar[:] = eval(self.Latitude)
        tvar = self.createVariable('time', 'd', ('site',))
        tvar.units = 'seconds since 1970-01-01 00:00:00+0000'
        tvar[:] = seconds

        self.description = ''.join(metalines)

    def avgSigma(self, vglvls=None, vgtop=None, hyai=None, hybi=None,
                 inplace=False, copyall=True, levelname='LAY'):
        """
        vglvls - sigma coordinate
        vgtop - top in Pa

        hyai - hybrid eta-component hPa
        hybi - hybrid sigma-component
        """
        if inplace:
            outf = self
        else:
            outf = self.copy(props=True, dimensions=True,
                             variables=copyall, data=copyall)
        myp = self.variables['Press'][:]  # hPa
        myo = self.variables['Ozone_mPa'][:] * 1e1  # mPa -> cPa
        psfc = myp[0, 0]
        if vglvls is not None:
            outpress_edges = vglvls * (psfc - vgtop / 100) + vgtop / 100
            nlvls = vglvls.size - 1
        elif hyai is not None:
            outpress_edges = hyai + hybi * psfc
            nlvls = hyai.size - 1
        outf.createDimension(levelname, nlvls)
        abovebottom = myp[0, :, None] <= outpress_edges[None, :-1]
        belowtop = myp[0, :, None] > outpress_edges[None, 1:]
        invglvl = abovebottom & belowtop
        out = np.ma.masked_all((1, nlvls), dtype='d')
        for li in range(nlvls):
            idx = invglvl[:, li]
            if not idx.any():
                continue
            lps = myp[:, idx]
            lops = np.ma.masked_invalid(myo[:, idx])
            lopsmask = np.ma.getmaskarray(lops)
            lp = np.ma.masked_where(lopsmask, lps).sum(-1)
            lop = lops.sum(-1)
            out[:, li] = lop / lp
        outv = outf.createVariable(
            'O3', 'f', ('site', levelname), missing_value=-999)
        outv.units = 'ppm'
        outv.description = 'Pressure-weighted average vmr in ppm'
        outv[:] = np.ma.filled(out[:], outv.missing_value)
        return outf



from __future__ import print_function

import code
import atexit
import os

__doc__ = r"""
.. _dumper
:mod:`dumper` -- PseudoNetCDF dump module
============================================

.. module:: pncdump
   :platform: Unix, Windows
   :synopsis: Provides ncdump like functionaility for PseudoNetCDF
.. moduleauthor:: Barron Henderson <barronh@unc.edu>

"""

__all__ = ['PNCConsole']


class PNCConsole(code.InteractiveConsole):
    """
    The PNCConsole is designed to create a fast Integrated
    Development Environment for scientific analysis.
    """

    def __init__(self, locals=None, filename='<console>',
                 histfile=None):
        """
        locals - dictionary of local variables
        filename - filename for error printouts
        histfile - path for history
        """
        if histfile is None:
            histfile = os.path.expanduser("~/.pncload-history")
        code.InteractiveConsole.__init__(
            self, locals=locals, filename=filename)
        self.init_history(histfile)

    def init_history(self, histfile):
        """
        Prepare history for use from histfile (typically last session)
        """
        import readline
        readline.parse_and_bind("tab: complete")
        if hasattr(readline, "read_history_file"):
            try:
                readline.read_history_file(histfile)
                readline.set_history_length(1000)
            except IOError:
                pass
            atexit.register(self.save_history, histfile)

    def save_history(self, histfile):
        """
        Write history from session to disk
        """
        import readline
        readline.write_history_file(histfile)


_torem = ['(', ')', '[', ']', '{', '}', '@', ',', ':', '.', '`', '=', ';',
          '+=', '-=', '*=', '/=', '//=', '%=', '&=', '|=', '^=', '>>=',
          '<<=', '**=', '+', '-', '*', '**', '/', '//', '%', '<<', '>>',
          '&', '|', '^', '~', '<', '>', '<=', '>=', '==', '!=', '<>']


def _clean(p):
    """
    Remove operator symbols from file names
    """
    for d in _torem:
        p = os.path.basename(p).replace(d, '')
    return p


def createconsole(ifiles, options):
    """
    Use standard pncparse ifiles and options to create
    a working environment.

    1) Access files by a short name or indexed name (ifile%d)
    2) Access variables by name and file index (var_%d)
    3) Has access to pylab and pncview functions
    """
    console = PNCConsole()
    exec("from pylab import *", None, console.locals)
    exec("import pylab as pl", None, console.locals)
    exec("import matplotlib.pyplot as plt", None, console.locals)
    exec("import PseudoNetCDF as pnc", None, console.locals)
    exec("from PseudoNetCDF.pncview import *; interactive(True)",
         None, console.locals)
    exec("from PseudoNetCDF.coordutil import *; interactive(True)",
         None, console.locals)
    if hasattr(options, 'ipath'):
        ipath = [ipath for ipath in options.ipath]
        ipaths = [_clean(ipath) for ipath in options.ipath]
    else:
        ipath = []
        ipaths = []
    spaths = [ipath[:6] for ipath in ipaths]
    spathsoc = dict([(k, 0) for k in spaths])
    spathso = []
    npathso = []
    for filei, spath in enumerate(spaths):
        spathso.append(spath + '_' + str(spathsoc[spath]))
        spathsoc[spath] += 1
        npathso.append('ifile%d' % filei)
    for rpath, npath, spath in zip(ipath, npathso, spathso):
        print('# ' + spath + ' = ' + npath + ' = ' + rpath)
    epaths_files = enumerate(zip(ipaths, npathso, spathso, ifiles))
    for filei, (ipath, npath, spath, ifile) in epaths_files:
        console.locals[npath] = ifile
        try:
            exec(ipath + ' = ' + npath, None, console.locals)
        except Exception:
            pass
        try:
            exec(spath + ' = ' + npath, None, console.locals)
        except Exception:
            pass
        console.locals.update(
            dict([('%s_%d' % (k, filei), v)
                  for k, v in ifile.variables.items()]))
    console.locals['ifiles'] = ifiles
    return console


def main():
    from .pncparse import pncparse
    ifiles, options = pncparse(has_ofile=False, interactive=False)
    console = createconsole(ifiles, options)
    console.interact()


if __name__ == '__main__':
    main()



import numpy as np
import io
from PseudoNetCDF import PseudoNetCDFFile
from PseudoNetCDF.pncwarn import warn
from datetime import datetime

_units = dict(
    Pressure='hPa', O3PartialPressure='mPa',
    Temperature='C', WindSpeed='m/s', WindDirection='degrees from north',
    RelativeHumidity='%', LevelCode='?', Duration='s', GPHeight='m',
    SampleTemperature='C', Latitude='degrees_north', Longitude='degrees_east')


class woudcsonde(PseudoNetCDFFile):
    """wouldcsonde provides a NetCDF-like interface to csv files from the
    ozonesonde data distributed by the Would Ozone and Ultraviolet
    Radiation Data Center (wouldc)

    http://woudc.org/data/products/
    """

    def _addmetavars(self, myf, key=''):
        """Adds metavariables for next two lines"""
        try:
            import pandas as pd
        except Exception:
            raise ImportError(
                'woudcsonde requires pandas; ' +
                'install pandas (e.g., pip install pandas)')
        metalines = myf.readline()
        nmetalines = 2
        newblock = ''
        while True:
            nextline = myf.readline()
            nstrip = nextline.strip()
            nmetalines += 1
            if nstrip == '':
                break
            elif nstrip.startswith('*'):
                break
            elif nstrip.startswith('#PROFILE'):
                newblock = nstrip[1:] + '_'
                break
            elif nstrip.startswith('#'):
                nmetalines -= 1
                newblock = nstrip.upper()[1:] + '_'
                nlines, newblock = self._addmetavars(myf, key=newblock)
                nmetalines += nlines
                break
            else:
                metalines += nextline

        if key == 'TIMESTAMP_':
            dtype = 'S64'
        else:
            dtype = None
        try:
            locdata = pd.read_csv(io.BytesIO(
                (metalines).encode()), dtype=dtype)
        except Exception as e:
            warn(key + ': ' + str(e))
            return nmetalines
        if key in ('PLATFORM_', 'LOCATION_'):
            dim1 = 'site'
        else:
            dim1 = 'flight'
        for colkey in locdata.columns:  # dtype.names:
            val = locdata[colkey]
            varkey = key + colkey
            dt = val.dtype.char
            odt = {'O': 'c'}.get(dt, dt)
            dim = {'c': (dim1, 'STRLEN'), 'S': (
                dim1, 'STRLEN')}.get(odt, (dim1,))
            if val.size > 1:
                if not key[:-1] in self.dimensions:
                    self.createDimension(key[:-1], val.size)
                dim = dim + (key[:-1],)
            var = self.createVariable(varkey, odt, dim)
            var.units = _units.get(colkey, 'unknown')
            var.long_name = colkey
            if odt in ('c', 'S'):
                var[:] = np.array([val]).astype('S64').view('S1')
            else:
                var[:] = val
        return nmetalines, newblock

    def __init__(self, path, debug=False, na_values=['\s+', '*', '99999']):
        """
        path - path or file-like object.
        """
        try:
            import pandas as pd
        except Exception:
            raise ImportError(
                'woudcsonde requires pandas; ' +
                'install pandas (e.g., pip install pandas)')
        if hasattr(path, 'readline'):
            myf = path
        else:
            myf = open(path, 'r')

        meta = ''
        li = 0
        self.createDimension('site', 1)
        self.createDimension('flight', 1)
        self.createDimension('STRLEN', 64)
        for l in myf:
            key = l.strip().upper()
            if key == '#PROFILE':
                break
            elif key.startswith('#'):
                if debug:
                    print(key)
                lines, newblock = self._addmetavars(myf, key[1:] + '_')
                li += lines
                if newblock == 'PROFILE_':
                    li -= 1
                    break
            else:
                meta += l
                li += 1

        # after #PROFILE loop is broken
        myf.close()
        readopts = dict(sep='\s*,', skiprows=li + 1,
                        engine='python', na_values=na_values, comment='*')
        data = pd.read_csv(path, **readopts)
        # indkey = data.columns[0]
        indkey = 'level'
        self.createDimension(indkey, len(data))
        self.metadata = meta
        for key in data.columns:
            try:
                values = data[key]
                var = self.createVariable(key, 'f', ('flight', indkey,))
                var.units = _units.get(key, 'unknown')
                var.long_name = key
                var[:] = values
            except Exception as e:
                warn(str(e) + '; ' + key + ' will not be written')

        datebytes = self.variables['TIMESTAMP_Date'].view('S64')[0, 0]
        datestr = datebytes.decode().strip()
        date = '-'.join(['%02d' % int(v) for v in datestr.split('-')])
        timebytes = self.variables['TIMESTAMP_Time'].view('S64')[0, 0]
        timestr = timebytes.decode().strip()
        time = (timestr + ':00')[:8]
        if time == ':00':
            time = '00:00:00'
        z = self.variables['TIMESTAMP_UTCOffset'].view(
            'S64')[0, 0].decode().strip()
        if z.startswith('+'):
            z = z[:3] + '00'
        elif z.startswith('-'):
            z = z[:3] + '00'
        else:
            z = '+' + z[:2] + '00'
        datestr = '{} {}{}'.format(date, time, z)
        rdatestr = '1970-01-01 00:00:00+0000'
        rdate = datetime.strptime(rdatestr, '%Y-%m-%d %H:%M:%S%z')
        outdate = datetime.strptime(datestr, '%Y-%m-%d %H:%M:%S%z')
        dt = (outdate - rdate).total_seconds()
        tvar = self.createVariable('time', 'd', ('flight',))
        tvar.long_name = 'time'
        tvar.units = 'seconds since ' + rdatestr
        tvar[...] = dt

    def avgSigma(self, vglvls=None, vgtop=None, hyai=None, hybi=None,
                 inplace=False, copyall=True, levelname='LAY'):
        """
        vglvls - sigma coordinate
        vgtop - top in Pa

        hyai - hybrid eta-component hPa
        hybi - hybrid sigma-component
        """
        if inplace:
            outf = self
        else:
            outf = self.copy(props=True, dimensions=True,
                             variables=copyall, data=copyall)
        myp = self.variables['Pressure'][:]  # hPa
        myo = self.variables['O3PartialPressure'][:] * 1e1  # mPa -> cPa
        psfc = myp[0, 0]
        if vglvls is not None:
            outpress_edges = vglvls * (psfc - vgtop / 100) + vgtop / 100
            nlvls = vglvls.size - 1
        elif hyai is not None:
            outpress_edges = hyai + hybi * psfc
            nlvls = hyai.size - 1
        outf.createDimension(levelname, nlvls)
        abovebottom = myp[0, :, None] <= outpress_edges[None, :-1]
        belowtop = myp[0, :, None] > outpress_edges[None, 1:]
        invglvl = abovebottom & belowtop
        out = np.ma.masked_all((1, nlvls), dtype='d')
        for li in range(nlvls):
            idx = invglvl[:, li]
            if not idx.any():
                continue
            lps = myp[:, idx]
            lops = np.ma.masked_invalid(myo[:, idx])
            lopsmask = np.ma.getmaskarray(lops)
            lp = np.ma.masked_where(lopsmask, lps).sum(-1)
            lop = lops.sum(-1)
            out[:, li] = lop / lp
        outv = outf.createVariable(
            'O3', 'f', ('flight', levelname), fill_value=-999)
        outv.units = 'ppm'
        outv.missing_value = -999
        outv.description = 'Pressure-weighted average vmr in ppm'
        outv[:] = out[:]
        return outf


if __name__ == "__main__":
    f = woudcsonde('20150813.ECC.1Z.1Z27925.JMA.csv')



__doc__ = """
.. _finst
:mod:`finst` -- Fine UAM-IV File Interfaces
===========================================

.. module:: finst
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map and random access read
   based file interfaces for CAMx fine UAM-IV files.
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
__all__ = ['Memmap']

from . import Memmap


if __name__ == '__main__':
    from PseudoNetCDF.camxfiles.finst.Memmap import finst
    from PseudoNetCDF.pncdump import pncdump_parser, \
        dump_from_cmd_line
    parser = pncdump_parser()
    parser.add_argument("cols", int)
    parser.add_argument("rows", int)
    (file_path, options, extra_args_dict) = parser.parse_args()

    dump_from_cmd_line(file_path, options, lambda path: finst)



import numpy as np
import sys
from PseudoNetCDF import PseudoNetCDFFile


def raob(inpath):
    nfields, = np.fromfile(inpath, dtype='>i4', count=1) / 4.
    fblock = np.fromfile(inpath, dtype='>i4')
    block = fblock[1:-2]
    block = block.reshape(-1, nfields + 1)[:, :-1]
    names, units = np.char.strip(block[:2].view('S4'))
    datas = (np.ma.masked_values(
        block[3, :], -2139062144).filled(-999) / 10.**block[2])
    data = (np.ma.masked_values(
        block[4:, :], -2139062144).filled(-999) / 10.**block[2])
    outf = PseudoNetCDFFile()
    outf.createDimension('level', data.shape[0])
    for k, u, v, vs in zip(names, units, data.T, datas):
        outf.createVariable(k, 'f', ('level',), units=u, values=v)
        outf.createVariable(k + '_FIRST', 'f', ('level',), units=u, values=vs)

    # np.savetxt(sys.stdout, (names, units), delimiter = ', ', fmt = '%7s')
    # np.savetxt(sys.stdout, data, delimiter = ', ', fmt = '%7.1f')
    return outf


if __name__ == '__main__':
    ifile0 = raob(sys.argv[1])



__all__ = [
    'bpch', 'gcnc', 'geos', 'bpch1', 'bpch2', 'ncf2bpch', 'cspec',
    'flightlogs'
]
from ._bpch import bpch1, ncf2bpch
from ._gcnc import gcnc
from ._newbpch import bpch2
from ._bpchmaster import bpch
from ._geos import geos
from ._cspec import cspec
from ._planelog import flightlogs



#!/usr/bin/env python
from __future__ import print_function

import os

from datetime import datetime
import argparse

from netCDF4 import Dataset
import numpy as np

from PseudoNetCDF.epafiles import aqsraw
from PseudoNetCDF.sci_var import stack_files
from PseudoNetCDF.pncgen import pncgen


def getbdate(x):
    return datetime.strptime(x, '%Y-%m-%d')


def getedate(x):
    return datetime.strptime(x + ' 23:59', '%Y-%m-%d %H:%M')


def getrdate(x):
    return datetime.strptime(x, '%Y-%m-%d %H:%M:%S')


parser = argparse.ArgumentParser(description="""Converts AQS Raw Hourly files for comparison with pncgen --extract files.

Example Workflow:
    $ %s --start-date=2006-08-01 --end-date=2012-08-01 --param=44201 GRIDCRO2D_Benchmark
    $ pncdump --header AQS_DATA_20060801-20060801.nc | grep lonlatcoords
            :lonlatcoords = "-87.881412,30.498001/-85.802182,33.281261/..." ;
    $ pncgen -s LAY,0 --extract="-87.881412,30.498001/-85.802182,33.281261/..." CCTM_V5g_par_Linux2_x86_64gfort.ACONC.CMAQ-BENCHMARK_20060801 Benchmark_20060801-20060801.nc
    $ pnceval AQS_DATA_20060801-20060801.nc Benchmark_20060801-20060801.nc
    """, formatter_class=argparse.RawDescriptionHelpFormatter)
parser.add_argument('-v', '--verbose', action='count',
                    dest='verbose', default=0)
parser.add_argument('--sampleval', default=None,
                    help='Defaults to "Sample Measurement" for hourly and "Arithmetic Mean" for daily')
parser.add_argument('--timeresolution',
                    choices=['daily', 'hourly'], default='hourly', help='Defaults to hourly')
parser.add_argument('-s', '--start-date', required=True, dest='bdate',
                    type=getbdate, help='Start date (inclusive) YYYY-MM-DD')
parser.add_argument('-e', '--end-date', required=True, dest='edate',
                    type=getedate, help='End date (inclusive) YYYY-MM-DD')
parser.add_argument('-r', '--ref-date', default=datetime(1900, 1, 1),
                    dest='rdate', type=getrdate, help='Reference date YYYYMMDD HH:MM:SS')
parser.add_argument('--param', type=str, default='44201',
                    nargs='?', help="Must exist as an AQS parameter")
spacegroup = parser.add_mutually_exclusive_group(required=False)
spacegroup.add_argument('--gridcro2d', dest='GRIDCRO2D',
                        help='CMAQ MCIP GRIDCRO2D file or any file that has LAT and LON variables')
spacegroup.add_argument('--wktpolygon')
parser.add_argument('-o', '--output', type=str, dest='outpath', nargs='?',
                    help='Path for output defaults to AQS_DATA_YYYYMMDD-YYYYMMDD.nc.')
parser.add_argument('-O', '--overwrite', dest='overwrite', default=False,
                    action='store_true', help='Ovewrite if output already exists.')

args = parser.parse_args()
if args.sampleval is None:
    if args.timeresolution == 'hourly':
        args.sampleval = "Sample Measurement"
    elif args.timeresolution == 'daily':
        args.sampleval = "Arithmetic Mean"
    else:
        raise KeyError()

if args.outpath is None:
    args.outpath = 'AQS_DATA_%s-%s.nc' % (
        args.bdate.strftime('%Y%m%d'), args.edate.strftime('%Y%m%d'))

if os.path.exists(args.outpath) and not args.overwrite:
    raise IOError('Path already exists: %s' % args.outpath)

if args.GRIDCRO2D is not None:
    f = Dataset(args.GRIDCRO2D)
    lon = f.variables['LON'][0, 0]
    lat = f.variables['LAT'][0, 0]
    args.minlon = llcrnrlon = lon[:, 0].max()
    args.maxlon = urcrnrlon = lon[:, -1].min()
    args.minlat = llcrnrlat = lat[0, :].max()
    args.maxlat = urcrnrlat = lat[-1, :].min()
    ll = (args.minlon, args.minlat)
    lr = (args.maxlon, args.minlat)
    ur = (args.maxlon, args.maxlat)
    ul = (args.minlon, args.maxlat)
    args.wktpolygon = "POLYGON ((%s %s, %s %s, %s %s, %s %s, %s %s))" % (
        ll + lr + ur + ul + ll)

years = np.arange(args.bdate.year, args.edate.year + 1)
yearpaths = []
for year in years:
    yearpath = '%s_%s_%s.csv' % (args.timeresolution, args.param, year)
    filepath = '%s_%s_%s.zip' % (args.timeresolution, args.param, year)
    print('Downloading', yearpath)
    url = 'http://aqsdr1.epa.gov/aqsweb/aqstmp/airdata/%s' % (filepath,)
    if os.path.exists(yearpath):
        print('..Already have ' + yearpath)
    else:
        if os.path.exists(filepath):
            print('..Already have ' + filepath)
        else:
            from urllib.request import urlopen, Request
            req = Request(url)
            ret = urlopen(req)
            data = ret.read()
            zipout = open(filepath, 'wb')
            zipout.write(data)
            zipout.flush()
            zipout.close()

        import zipfile
        zf = zipfile.ZipFile(filepath)
        zf.extract(yearpath)
    yearpaths.append(yearpath)

infiles = []
for yearpath in yearpaths:
    infile = aqsraw(yearpath, timeresolution=args.timeresolution, bdate=args.bdate, edate=args.edate,
                    rdate=args.rdate, wktpolygon=args.wktpolygon, sampleval=args.sampleval, verbose=args.verbose)
    infiles.append(infile)

outfile = stack_files(infiles, 'time')
persisted = pncgen(outfile, args.outpath)
persisted.sync()
print('Successful')



__all__ = ['get_chemparam_names', 'get_uamiv_units']

_camx_units = {
    'EMISSIONS ': {True: 'g/time', False: 'mol/time'},
    'AVERAGE   ': {True: 'micrograms/m**3', False: 'ppm'},
    'BOUNDARY  ': {True: 'micrograms/m**3', False: 'ppm'},
    'INSTANT   ': {True: 'micrograms/m**3', False: 'micromoles/m**3'},
    'IPR': {True: 'micrograms/m**3', False: 'micromoles/m**3'},
    'AIRQUALITY': {True: 'micrograms/m**3', False: 'ppm'},
    'DEPOSITION': {True: 'g/ha', False: 'mol/ha'},
}


def get_chemparam_names(chemparampath):
    inlines = open(chemparampath, 'r').readlines()
    startaero = None
    for li, l in enumerate(inlines):
        if l[:14] == "     Gas Spec ":
            startgas = li + 1
        if l[:14] == "     Aero Spec":
            endgas = li
            startaero = li + 1
        if l[:16] == "Reaction Records":
            if startaero is None:
                endgas = li
            endaero = li
            break

    gaskeys = tuple([l[5:15].strip() for l in inlines[startgas:endgas]])
    if startaero is None:
        return dict(gas=gaskeys, aer=())
    aerkeys = tuple([l[5:15].strip() for l in inlines[startaero:endaero]])
    return dict(gas=gaskeys, aerosol=aerkeys)


def get_uamiv_units(filename, key, aerosol_names=None):
    if aerosol_names is None:
        from PseudoNetCDF.camxfiles.aerosol_names import aerosol_names
    if key[-3:] in ('_DV',):
        return 'm/s'
    elif key[-3:] in ('_DD', '_WD'):
        return _camx_units['DEPOSITION'][key[:-3] in aerosol_names]
    return _camx_units[filename][key in aerosol_names]



__all__ = ['add_cf_from_ioapi', 'get_ioapi_sphere', 'add_ioapi_from_ioapi',
           'add_ioapi_from_cf', 'getmapdef', 'add_cf_from_wrfioapi']


from ._ioapi import add_cf_from_ioapi, get_ioapi_sphere
from ._ioapi import add_ioapi_from_ioapi, add_ioapi_from_cf, getmapdef
from ._wrfioapi import add_cf_from_wrfioapi



__all__ = ['getwriterdict', 'registerwriter']

_writers = []


def testwriter(writer, *args, **kwds):
    try:
        writer(*args, **kwds)
        return True
    except Exception:
        return False


def registerwriter(name, writer):
    global _writers
    _writers.insert(0, (name, writer))


def getwriterdict():
    return dict(_writers)


def pncwrite(*args, **kwds):
    """See PseudoNetCDF.pncgen.pncgen

    *args : iterable
    **kwds : keywords
        keywords for pncgen
    help : boolean
        without format, returns help of pncopen and with format keyword,
        returns help of the function that writes that format.
    """
    from PseudoNetCDF.pncgen import pncgen
    formathelp = kwds.pop('help', False)
    if formathelp:
        if 'format' in kwds:
            format = kwds['format']
            writer = getwriterdict()[format]
            return help(writer)
        else:
            return help(pncgen)
    return pncgen(*args, **kwds)



__doc__ = """
.. _humidity
:mod:`humidity` -- Humidity File Interfaces
===========================================

.. module:: humidity
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map and random access read
   based file interfaces for CAMx humidity files.
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
__all__ = ['Memmap', 'Read', 'Transforms', 'Write']

from . import Memmap
from . import Read
from . import Write
from . import Transforms

if __name__ == '__main__':
    from PseudoNetCDF.camxfiles.humidity.Memmap import humidity
    from PseudoNetCDF.pncdump import pncdump_parser, \
        dump_from_cmd_line
    parser = pncdump_parser()
    parser.add_argument("cols", int)
    parser.add_argument("rows", int)
    (file_path, options, extra_args_dict) = parser.parse_args()

    dump_from_cmd_line(file_path, options, humidity)
